import { version, gStyle, httpRequest, createHttpRequest, loadScript, decodeUrl,
         source_dir, settings, internals, findFunction,
         isArrayProto, isRootCollection, isBatchMode, isNodeJs, _ensureJSROOT } from './core.mjs';

import { select as d3_select } from './d3.mjs';

import { openFile } from './io.mjs';

import { getRGBfromTColor } from './base/colors.mjs';

import { BasePainter, getElementRect } from './base/BasePainter.mjs';

import { ObjectPainter } from './base/ObjectPainter.mjs';

import { loadJSDOM, getElementMainPainter, cleanup } from './painter.mjs';

import { createMenu } from './gui/menu.mjs';

import { produceLegend } from './hist/TPavePainter.mjs';

import { getDrawSettings, getDrawHandle, canDrawHandle, addDrawFunc, draw, redraw } from './draw.mjs';

import { BatchDisplay, GridDisplay, FlexibleDisplay, BrowserLayout } from './gui/display.mjs';

import { showProgress, ToolbarIcons, registerForResize } from './gui/utils.mjs';

function canExpandHandle(handle) {
   return handle?.expand || handle?.get_expand || handle?.expand_item;
}

/** @summary draw list content
  * @desc used to draw all items from TList or TObjArray inserted into the TCanvas list of primitives
  * @private */
function drawList(dom, lst, opt) {
   if (!lst || !lst.arr)
      return Promise.resolve(null);

   let obj = {
     divid: dom,
     lst: lst,
     opt: opt,
     indx: -1,
     painter: null,
     draw_next: function() {
        while (++this.indx < this.lst.arr.length) {
           let item = this.lst.arr[this.indx],
               opt = (this.lst.opt && this.lst.opt[this.indx]) ? this.lst.opt[this.indx] : this.opt;
           if (!item) continue;
           return draw(this.getDom(), item, opt).then(p => {
              if (p && !this.painter) this.painter = p;
              return this.draw_next(); // reenter loop
           });
        }

        return Promise.resolve(this.painter);
     }
   }

   return obj.draw_next();
}

// ===================== hierarchy scanning functions ==================================

/** @summary Create hierarchy elements for TFolder object
  * @private */
function folderHierarchy(item, obj) {

   if (!obj || !('fFolders' in obj) || (obj.fFolders===null)) return false;

   if (obj.fFolders.arr.length===0) { item._more = false; return true; }

   item._childs = [];

   for ( let i = 0; i < obj.fFolders.arr.length; ++i) {
      let chld = obj.fFolders.arr[i];
      item._childs.push( {
         _name : chld.fName,
         _kind : "ROOT." + chld._typename,
         _obj : chld
      });
   }
   return true;
}

/** @summary Create hierarchy elements for TTask object
  * @private */
function taskHierarchy(item, obj) {
   // function can be used for different derived classes
   // we show not only child tasks, but all complex data members

   if (!obj || !('fTasks' in obj) || (obj.fTasks === null)) return false;

   objectHierarchy(item, obj, { exclude: ['fTasks', 'fName'] } );

   if ((obj.fTasks.arr.length===0) && (item._childs.length==0)) { item._more = false; return true; }

   // item._childs = [];

   for ( let i = 0; i < obj.fTasks.arr.length; ++i) {
      let chld = obj.fTasks.arr[i];
      item._childs.push({
         _name : chld.fName,
         _kind : "ROOT." + chld._typename,
         _obj : chld
      });
   }
   return true;
}

/** @summary Create hierarchy elements for TList object
  * @private */
function listHierarchy(folder, lst) {
   if (!isRootCollection(lst)) return false;

   if ((lst.arr === undefined) || (lst.arr.length === 0)) {
      folder._more = false;
      return true;
   }

   let do_context = false, prnt = folder;
   while (prnt) {
      if (prnt._do_context) do_context = true;
      prnt = prnt._parent;
   }

   // if list has objects with similar names, create cycle number for them
   let ismap = (lst._typename == 'TMap'), names = [], cnt = [], cycle = [];

   for (let i = 0; i < lst.arr.length; ++i) {
      let obj = ismap ? lst.arr[i].first : lst.arr[i];
      if (!obj) continue; // for such objects index will be used as name
      let objname = obj.fName || obj.name;
      if (!objname) continue;
      let indx = names.indexOf(objname);
      if (indx>=0) {
         cnt[indx]++;
      } else {
         cnt[names.length] = cycle[names.length] = 1;
         names.push(objname);
      }
   }

   folder._childs = [];
   for (let i = 0; i < lst.arr.length; ++i) {
      let obj = ismap ? lst.arr[i].first : lst.arr[i],
          item = !obj || !obj._typename ?
           {
            _name: i.toString(),
            _kind: "ROOT.NULL",
            _title: "NULL",
            _value: "null",
            _obj: null
          } : {
            _name: obj.fName || obj.name,
            _kind: "ROOT." + obj._typename,
            _title: (obj.fTitle || "") + " type:"  +  obj._typename,
            _obj: obj
        };

        switch(obj._typename) {
           case 'TColor': item._value = getRGBfromTColor(obj); break;
           case 'TText':
           case 'TLatex': item._value = obj.fTitle; break;
           case 'TObjString': item._value = obj.fString; break;
           default: if (lst.opt && lst.opt[i] && lst.opt[i].length) item._value = lst.opt[i];
        }

        if (do_context && canDrawHandle(obj._typename)) item._direct_context = true;

        // if name is integer value, it should match array index
        if (!item._name || (Number.isInteger(parseInt(item._name)) && (parseInt(item._name) !== i))
            || (lst.arr.indexOf(obj) < i)) {
           item._name = i.toString();
        } else {
           // if there are several such names, add cycle number to the item name
           let indx = names.indexOf(obj.fName);
           if ((indx >= 0) && (cnt[indx] > 1)) {
              item._cycle = cycle[indx]++;
              item._keyname = item._name;
              item._name = item._keyname + ";" + item._cycle;
           }
        }

      folder._childs.push(item);
   }
   return true;
}

/** @summary Create hierarchy of TKey lists in file or sub-directory
  * @private */
function keysHierarchy(folder, keys, file, dirname) {

   if (keys === undefined) return false;

   folder._childs = [];

   for (let i = 0; i < keys.length; ++i) {
      let key = keys[i],
          item = {
         _name: key.fName + ";" + key.fCycle,
         _cycle: key.fCycle,
         _kind: "ROOT." + key.fClassName,
         _title: key.fTitle,
         _keyname: key.fName,
         _readobj: null,
         _parent: folder
      };

      if (key.fObjlen > 1e5) item._title += ' (size: ' + (key.fObjlen/1e6).toFixed(1) + 'MB)';

      if ('fRealName' in key)
         item._realname = key.fRealName + ";" + key.fCycle;

      if (key.fClassName == 'TDirectory' || key.fClassName == 'TDirectoryFile') {
         let dir = null;
         if (dirname && file) dir = file.getDir(dirname + key.fName);
         if (!dir) {
            item._more = true;
            item._expand = function(node, obj) {
               // one can get expand call from child objects - ignore them
               return keysHierarchy(node, obj.fKeys);
            };
         } else {
            // remove cycle number - we have already directory
            item._name = key.fName;
            keysHierarchy(item, dir.fKeys, file, dirname + key.fName + "/");
         }
      } else if ((key.fClassName == 'TList') && (key.fName == 'StreamerInfo')) {
         if (settings.SkipStreamerInfos) continue;
         item._name = 'StreamerInfo';
         item._kind = "ROOT.TStreamerInfoList";
         item._title = "List of streamer infos for binary I/O";
         item._readobj = file.fStreamerInfos;
      }

      folder._childs.push(item);
   }

   return true;
}

/** @summary Create hierarchy for arbitrary object
  * @private */
function objectHierarchy(top, obj, args = undefined) {
   if (!top || (obj===null)) return false;

   top._childs = [];

   let proto = Object.prototype.toString.apply(obj);

   if (proto === '[object DataView]') {

      let item = {
          _parent: top,
          _name: 'size',
          _value: obj.byteLength.toString(),
          _vclass: 'h_value_num'
      };

      top._childs.push(item);
      let namelen = (obj.byteLength < 10) ? 1 : Math.log10(obj.byteLength);

      for (let k = 0; k < obj.byteLength; ++k) {
         if (k % 16 === 0) {
            item = {
              _parent: top,
              _name: k.toString(),
              _value: "",
              _vclass: 'h_value_num'
            };
            while (item._name.length < namelen) item._name = "0" + item._name;
            top._childs.push(item);
         }

         let val = obj.getUint8(k).toString(16);
         while (val.length<2) val = "0"+val;
         if (item._value.length>0)
            item._value += (k%4===0) ? " | " : " ";

         item._value += val;
      }
      return true;
   }

   // check nosimple property in all parents
   let nosimple = true, do_context = false, prnt = top;
   while (prnt) {
      if (prnt._do_context) do_context = true;
      if ('_nosimple' in prnt) { nosimple = prnt._nosimple; break; }
      prnt = prnt._parent;
   }

   let isarray = (isArrayProto(proto) > 0) && obj.length,
       compress = isarray && (obj.length > settings.HierarchyLimit),  arrcompress = false;

   if (isarray && (top._name==="Object") && !top._parent) top._name = "Array";

   if (compress) {
      arrcompress = true;
      for (let k=0;k<obj.length;++k) {
         let typ = typeof obj[k];
         if ((typ === 'number') || (typ === 'boolean') || (typ=='string' && (obj[k].length<16))) continue;
         arrcompress = false; break;
      }
   }

   if (!('_obj' in top))
      top._obj = obj;
   else if (top._obj !== obj)
      alert('object missmatch');

   if (!top._title) {
      if (obj._typename)
         top._title = "ROOT." + obj._typename;
      else if (isarray)
         top._title = "Array len: " + obj.length;
   }

   if (arrcompress) {
      for (let k = 0; k < obj.length;) {

         let nextk = Math.min(k+10,obj.length), allsame = true, prevk = k;

         while (allsame) {
            allsame = true;
            for (let d=prevk;d<nextk;++d)
               if (obj[k]!==obj[d]) allsame = false;

            if (allsame) {
               if (nextk===obj.length) break;
               prevk = nextk;
               nextk = Math.min(nextk+10,obj.length);
            } else if (prevk !== k) {
               // last block with similar
               nextk = prevk;
               allsame = true;
               break;
            }
         }

         let item = { _parent: top, _name: k+".."+(nextk-1), _vclass: 'h_value_num' };

         if (allsame) {
            item._value = obj[k].toString();
         } else {
            item._value = "";
            for (let d=k;d<nextk;++d)
               item._value += ((d===k) ? "[ " : ", ") + obj[d].toString();
            item._value += " ]";
         }

         top._childs.push(item);

         k = nextk;
      }
      return true;
   }

   let lastitem, lastkey, lastfield, cnt;

   for (let key in obj) {
      if ((key == '_typename') || (key[0]=='$')) continue;
      let fld = obj[key];
      if (typeof fld == 'function') continue;
      if (args?.exclude && (args.exclude.indexOf(key) >= 0)) continue;

      if (compress && lastitem) {
         if (lastfield===fld) { ++cnt; lastkey = key; continue; }
         if (cnt > 0) lastitem._name += ".." + lastkey;
      }

      let item = { _parent: top, _name: key };

      if (compress) { lastitem = item;  lastkey = key; lastfield = fld; cnt = 0; }

      if (fld === null) {
         item._value = item._title = "null";
         if (!nosimple) top._childs.push(item);
         continue;
      }

      let simple = false;

      if (typeof fld == 'object') {

         proto = Object.prototype.toString.apply(fld);

         if (isArrayProto(proto) > 0) {
            item._title = "array len=" + fld.length;
            simple = (proto != '[object Array]');
            if (fld.length === 0) {
               item._value = "[ ]";
               item._more = false; // hpainter will not try to expand again
            } else {
               item._value = "[...]";
               item._more = true;
               item._expand = objectHierarchy;
               item._obj = fld;
            }
         } else if (proto === "[object DataView]") {
            item._title = 'DataView len=' + fld.byteLength;
            item._value = "[...]";
            item._more = true;
            item._expand = objectHierarchy;
            item._obj = fld;
         } else if (proto === "[object Date]") {
            item._more = false;
            item._title = 'Date';
            item._value = fld.toString();
            item._vclass = 'h_value_num';
         } else {

            if (fld.$kind || fld._typename)
               item._kind = item._title = "ROOT." + (fld.$kind || fld._typename);

            if (fld._typename) {
               item._title = fld._typename;
               if (do_context && canDrawHandle(fld._typename)) item._direct_context = true;
            }

            // check if object already shown in hierarchy (circular dependency)
            let curr = top, inparent = false;
            while (curr && !inparent) {
               inparent = (curr._obj === fld);
               curr = curr._parent;
            }

            if (inparent) {
               item._value = "{ prnt }";
               item._vclass = 'h_value_num';
               item._more = false;
               simple = true;
            } else {
               item._obj = fld;
               item._more = false;

               switch(fld._typename) {
                  case 'TColor': item._value = getRGBfromTColor(fld); break;
                  case 'TText':
                  case 'TLatex': item._value = fld.fTitle; break;
                  case 'TObjString': item._value = fld.fString; break;
                  default:
                     if (isRootCollection(fld) && (typeof fld.arr === "object")) {
                        item._value = fld.arr.length ? "[...]" : "[]";
                        item._title += ", size:"  + fld.arr.length;
                        if (fld.arr.length>0) item._more = true;
                     } else {
                        item._more = true;
                        item._value = "{ }";
                     }
               }
            }
         }
      } else if ((typeof fld === 'number') || (typeof fld === 'boolean')) {
         simple = true;
         if (key == 'fBits')
            item._value = "0x" + fld.toString(16);
         else
            item._value = fld.toString();
         item._vclass = 'h_value_num';
      } else if (typeof fld === 'string') {
         simple = true;
         item._value = '&quot;' + fld.replace(/\&/g, '&amp;').replace(/\"/g, '&quot;').replace(/</g, '&lt;').replace(/>/g, '&gt;') + '&quot;';
         item._vclass = 'h_value_str';
      } else if (typeof fld === 'undefined') {
         simple = true;
         item._value = "undefined";
         item._vclass = 'h_value_num';
      } else {
         simple = true;
         alert('miss ' + key + '  ' + typeof fld);
      }

      if (!simple || !nosimple)
         top._childs.push(item);
   }

   if (compress && lastitem && (cnt>0)) lastitem._name += ".." + lastkey;

   return true;
}

/** @summary Create hierarchy for streamer info object
  * @private */
function createStreamerInfoContent(lst) {
   let h = { _name : "StreamerInfo", _childs : [] };

   for (let i = 0; i < lst.arr.length; ++i) {
      let entry = lst.arr[i];

      if (entry._typename == "TList") continue;

      if (typeof entry.fName == 'undefined') {
         console.warn(`strange element in StreamerInfo with type ${entry._typename}`);
         continue;
      }

      let item = {
         _name : entry.fName + ";" + entry.fClassVersion,
         _kind : "class " + entry.fName,
         _title : "class:" + entry.fName + ' version:' + entry.fClassVersion + ' checksum:' + entry.fCheckSum,
         _icon: "img_class",
         _childs : []
      };

      if (entry.fTitle != '') item._title += '  ' + entry.fTitle;

      h._childs.push(item);

      if (typeof entry.fElements == 'undefined') continue;
      for ( let l = 0; l < entry.fElements.arr.length; ++l) {
         let elem = entry.fElements.arr[l];
         if (!elem || !elem.fName) continue;
         let info = elem.fTypeName + " " + elem.fName,
             title = elem.fTypeName + " type:" + elem.fType;
         if (elem.fArrayDim===1)
            info += "[" + elem.fArrayLength + "]";
         else
            for (let dim = 0; dim < elem.fArrayDim; ++dim)
               info += "[" + elem.fMaxIndex[dim] + "]";
         if (elem.fBaseVersion === 4294967295) info += ":-1"; else
         if (elem.fBaseVersion !== undefined) info += ":" + elem.fBaseVersion;
         info += ";";
         if (elem.fTitle) info += " // " + elem.fTitle;

         item._childs.push({ _name : info, _title: title, _kind: elem.fTypeName, _icon: (elem.fTypeName == 'BASE') ? "img_class" : "img_member" });
      }
      if (!item._childs.length) delete item._childs;
   }

   return h;
}

/** @summary Create hierarchy for object inspector
  * @private */
function createInspectorContent(obj) {
   let h = { _name: "Object", _title: "", _click_action: "expand", _nosimple: false, _do_context: true };

   if ((typeof obj.fName === 'string') && (obj.fName.length > 0))
      h._name = obj.fName;

   if ((typeof obj.fTitle === 'string') && (obj.fTitle.length > 0))
      h._title = obj.fTitle;

   if (obj._typename)
      h._title += "  type:" + obj._typename;

   if (isRootCollection(obj)) {
      h._name = obj.name || obj._typename;
      listHierarchy(h, obj);
   } else {
      objectHierarchy(h, obj);
   }

   return h;
}


/** @summary Parse string value as array.
 * @desc It could be just simple string:  "value" or
 * array with or without string quotes:  [element], ['elem1',elem2]
 * @private */
let parseAsArray = val => {

   let res = [];

   if (typeof val != 'string') return res;

   val = val.trim();
   if (val=="") return res;

   // return as array with single element
   if ((val.length < 2) || (val[0] != '[') || (val[val.length-1] != ']')) {
      res.push(val); return res;
   }

   // try to split ourself, checking quotes and brackets
   let nbr = 0, nquotes = 0, ndouble = 0, last = 1;

   for (let indx = 1; indx < val.length; ++indx) {
      if (nquotes > 0) {
         if (val[indx]==="'") nquotes--;
         continue;
      }
      if (ndouble > 0) {
         if (val[indx]==='"') ndouble--;
         continue;
      }
      switch (val[indx]) {
         case "'": nquotes++; break;
         case '"': ndouble++; break;
         case "[": nbr++; break;
         case "]": if (indx < val.length - 1) { nbr--; break; }
         case ",":
            if (nbr === 0) {
               let sub = val.substring(last, indx).trim();
               if ((sub.length > 1) && (sub[0] == sub[sub.length-1]) && ((sub[0] == '"') || (sub[0] == "'")))
                  sub = sub.substr(1, sub.length-2);
               res.push(sub);
               last = indx+1;
            }
            break;
      }
   }

   if (res.length === 0)
      res.push(val.substr(1, val.length-2).trim());

   return res;
}



/** @summary central function for expand of all online items
  * @private */
function onlineHierarchy(node, obj) {
   if (obj && node && ('_childs' in obj)) {

      for (let n = 0; n < obj._childs.length; ++n)
         if (obj._childs[n]._more || obj._childs[n]._childs)
            obj._childs[n]._expand = onlineHierarchy;

      node._childs = obj._childs;
      obj._childs = null;
      return true;
   }

   return false;
}

// ==============================================================

/** @summary Current hierarchy painter
  * @desc Instance of {@link HierarchyPainter} object
  * @private */
let first_hpainter = null;

function getHPainter() { return first_hpainter; }

/**
  * @summary Painter of hierarchical structures
  *
  * @example
  * // create hierarchy painter in "myTreeDiv"
  * let h = new HierarchyPainter("example", "myTreeDiv");
  * // configure 'simple' layout in "myMainDiv"
  * // one also can specify "grid2x2" or "flex" or "tabs"
  * h.setDisplay("simple", "myMainDiv");
  * // open file and display element
  * h.openRootFile("https://root.cern/js/files/hsimple.root").then(() => h.display("hpxpy;1","colz")); */

class HierarchyPainter extends BasePainter {

   /** @summary Create painter
     * @param {string} name - symbolic name
     * @param {string} frameid - element id where hierarchy is drawn
     * @param {string} [backgr] - background color */
   constructor(name, frameid, backgr) {
      super(frameid);
      this.name = name;
      this.h = null; // hierarchy
      this.with_icons = true;
      this.background = backgr;
      this.files_monitoring = !frameid; // by default files monitored when nobrowser option specified
      this.nobrowser = (frameid === null);

      // remember only very first instance
      if (!first_hpainter)
         first_hpainter = this;
   }

   /** @summary Cleanup hierarchy painter
     * @desc clear drawing and browser */
   cleanup() {
      this.clearHierarchy(true);

      super.cleanup();

      if (first_hpainter === this)
         first_hpainter = null;
   }

   /** @summary Create file hierarchy
     * @private */
   fileHierarchy(file) {
      let painter = this;

      let folder = {
         _name : file.fFileName,
         _title : (file.fTitle ? (file.fTitle + ", path ") : "")  + file.fFullURL,
         _kind : "ROOT.TFile",
         _file : file,
         _fullurl : file.fFullURL,
         _localfile : file.fLocalFile,
         _had_direct_read : false,
         // this is central get method, item or itemname can be used, returns promise
         _get : function(item, itemname) {

            let fff = this; // file item

            if (item && item._readobj)
               return Promise.resolve(item._readobj);

            if (item) itemname = painter.itemFullName(item, fff);

            function ReadFileObject(file) {
               if (!fff._file) fff._file = file;

               if (!file) return Promise.resolve(null);

               return file.readObject(itemname).then(obj => {

                  // if object was read even when item did not exist try to reconstruct new hierarchy
                  if (!item && obj) {
                     // first try to found last read directory
                     let d = painter.findItem({name:itemname, top:fff, last_exists:true, check_keys:true });
                     if ((d!=null) && ('last' in d) && (d.last!=fff)) {
                        // reconstruct only subdir hierarchy
                        let dir = file.getDir(painter.itemFullName(d.last, fff));
                        if (dir) {
                           d.last._name = d.last._keyname;
                           let dirname = painter.itemFullName(d.last, fff);
                           keysHierarchy(d.last, dir.fKeys, file, dirname + "/");
                        }
                     } else {
                        // reconstruct full file hierarchy
                        keysHierarchy(fff, file.fKeys, file, "");
                     }
                     item = painter.findItem({name:itemname, top: fff});
                  }

                  if (item) {
                     item._readobj = obj;
                     // remove cycle number for objects supporting expand
                     if ('_expand' in item) item._name = item._keyname;
                  }

                  return obj;
               });
            }

            if (fff._file) return ReadFileObject(fff._file);
            if (fff._localfile) return openFile(fff._localfile).then(f => ReadFileObject(f));
            if (fff._fullurl) return openFile(fff._fullurl).then(f => ReadFileObject(f));
            return Promise.resolve(null);
         }
      };

      keysHierarchy(folder, file.fKeys, file, "");

      return folder;
   }

   /** @summary Iterate over all items in hierarchy
     * @param {function} func - function called for every item
     * @param {object} [top] - top item to start from
     * @private */
   forEachItem(func, top) {
      function each_item(item, prnt) {
         if (!item) return;
         if (prnt) item._parent = prnt;
         func(item);
         if ('_childs' in item)
            for (let n = 0; n < item._childs.length; ++n)
               each_item(item._childs[n], item);
      }

      if (typeof func == 'function')
         each_item(top || this.h);
   }

   /** @summary Search item in the hierarchy
     * @param {object|string} arg - item name or object with arguments
     * @param {string} arg.name -  item to search
     * @param {boolean} [arg.force] - specified elements will be created when not exists
     * @param {boolean} [arg.last_exists] -  when specified last parent element will be returned
     * @param {boolean} [arg.check_keys] - check TFile keys with cycle suffix
     * @param {boolean} [arg.allow_index] - let use sub-item indexes instead of name
     * @param {object} [arg.top] - element to start search from
     * @private */
   findItem(arg) {

      function find_in_hierarchy(top, fullname) {

         if (!fullname || (fullname.length == 0) || !top) return top;

         let pos = fullname.length;

         if (!top._parent && (top._kind !== 'TopFolder') && (fullname.indexOf(top._name)===0)) {
            // it is allowed to provide item name, which includes top-parent like file.root/folder/item
            // but one could skip top-item name, if there are no other items
            if (fullname === top._name) return top;

            let len = top._name.length;
            if (fullname[len] == "/") {
               fullname = fullname.substr(len+1);
               pos = fullname.length;
            }
         }

         function process_child(child, ignore_prnt) {
            // set parent pointer when searching child
            if (!ignore_prnt) child._parent = top;

            if ((pos >= fullname.length-1) || (pos < 0)) return child;

            return find_in_hierarchy(child, fullname.substr(pos + 1));
         }

         while (pos > 0) {
            // we try to find element with slashes inside - start from full name
            let localname = (pos >= fullname.length) ? fullname : fullname.substr(0, pos);

            if (top._childs) {
               // first try to find direct matched item
               for (let i = 0; i < top._childs.length; ++i)
                  if (top._childs[i]._name == localname)
                     return process_child(top._childs[i]);

               // if first child online, check its elements
               if ((top._kind === 'TopFolder') && (top._childs[0]._online!==undefined))
                  for (let i = 0; i < top._childs[0]._childs.length; ++i)
                     if (top._childs[0]._childs[i]._name == localname)
                        return process_child(top._childs[0]._childs[i], true);

               // if allowed, try to found item with key
               if (arg.check_keys) {
                  let newest = null;
                  for (let i = 0; i < top._childs.length; ++i) {
                    if (top._childs[i]._keyname === localname) {
                       if (!newest || (newest._cycle < top._childs[i]._cycle)) newest = top._childs[i];
                    }
                  }
                  if (newest) return process_child(newest);
               }

               let allow_index = arg.allow_index;
               if ((localname[0] === '[') && (localname[localname.length-1] === ']') &&
                    /^\d+$/.test(localname.substr(1,localname.length-2))) {
                  allow_index = true;
                  localname = localname.substr(1,localname.length-2);
               }

               // when search for the elements it could be allowed to check index
               if (allow_index && /^\d+$/.test(localname)) {
                  let indx = parseInt(localname);
                  if (Number.isInteger(indx) && (indx >= 0) && (indx < top._childs.length))
                     return process_child(top._childs[indx]);
               }
            }

            pos = fullname.lastIndexOf("/", pos - 1);
         }

         if (arg.force) {
             // if did not found element with given name we just generate it
             if (top._childs === undefined) top._childs = [];
             pos = fullname.indexOf("/");
             let child = { _name: ((pos < 0) ? fullname : fullname.substr(0, pos)) };
             top._childs.push(child);
             return process_child(child);
         }

         return (arg.last_exists && top) ? { last: top, rest: fullname } : null;
      }

      let top = this.h, itemname = "";

      if (arg === null) return null; else
      if (typeof arg == 'string') { itemname = arg; arg = {}; } else
      if (typeof arg == 'object') { itemname = arg.name; if ('top' in arg) top = arg.top; } else
         return null;

      if (itemname === "__top_folder__") return top;

      if ((typeof itemname == 'string') && (itemname.indexOf("img:")==0)) return null;

      return find_in_hierarchy(top, itemname);
   }

   /** @summary Produce full string name for item
     * @param {Object} node - item element
     * @param {Object} [uptoparent] - up to which parent to continue
     * @param {boolean} [compact] - if specified, top parent is not included
     * @returns {string} produced name
     * @private */
   itemFullName(node, uptoparent, compact) {

      if (node && node._kind ==='TopFolder') return "__top_folder__";

      let res = "";

      while (node) {
         // online items never includes top-level folder
         if ((node._online!==undefined) && !uptoparent) return res;

         if ((node === uptoparent) || (node._kind==='TopFolder')) break;
         if (compact && !node._parent) break; // in compact form top-parent is not included
         if (res.length > 0) res = "/" + res;
         res = node._name + res;
         node = node._parent;
      }

      return res;
   }

    /** @summary Executes item marked as 'Command'
      * @desc If command requires additional arguments, they could be specified as extra arguments arg1, arg2, ...
      * @param {String} itemname - name of command item
      * @param {Object} [elem] - HTML element for command execution
      * @param [arg1] - first optional argument
      * @param [arg2] - second optional argument and so on
      * @returns {Promise} with command result */
   executeCommand(itemname, elem) {

      let hitem = this.findItem(itemname),
          url = this.getOnlineItemUrl(hitem) + "/cmd.json",
          d3node = d3_select(elem),
          cmdargs = [];

      if ('_numargs' in hitem)
         for (let n = 0; n < hitem._numargs; ++n)
            cmdargs.push((n+2 < arguments.length) ? arguments[n+2] : "");

      let promise = (cmdargs.length == 0) || !elem ? Promise.resolve(cmdargs) :
                     createMenu().then(menu => menu.showCommandArgsDialog(hitem._name, cmdargs));

      return promise.then(args => {
         if (args === null) return false;

         let urlargs = "";
         for (let k = 0; k < args.length; ++k)
            urlargs += `${k > 0 ?  "&" : "?"}arg${k+1}=${args[k]}`;

        if (!d3node.empty()) {
            d3node.style('background','yellow');
            if (hitem && hitem._title) d3node.attr('title', "Executing " + hitem._title);
         }

         return httpRequest(url + urlargs, 'text').then(res => {
            if (!d3node.empty()) {
               let col = ((res != null) && (res != 'false')) ? 'green' : 'red';
               if (hitem && hitem._title) d3node.attr('title', hitem._title + " lastres=" + res);
               d3node.style('background', col);
               setTimeout(() => d3node.style('background', ''), 2000);
               if ((col == 'green') && ('_hreload' in hitem))
                  this.reload();
               if ((col == 'green') && ('_update_item' in hitem))
                  this.updateItems(hitem._update_item.split(";"));
            }
            return res;
         });
      });
   }

   /** @summary Get object item with specified name
     * @desc depending from provided option, same item can generate different object types
     * @param {Object} arg - item name or config object
     * @param {string} arg.name - item name
     * @param {Object} arg.item - or item itself
     * @param {string} options - supposed draw options
     * @returns {Promise} with object like { item, obj, itemname }
     * @private */
   getObject(arg, options) {

      let itemname, item, result = { item: null, obj: null };

      if (arg===null) return Promise.resolve(result);

      if (typeof arg === 'string') {
         itemname = arg;
      } else if (typeof arg === 'object') {
         if ((arg._parent!==undefined) && (arg._name!==undefined) && (arg._kind!==undefined)) item = arg; else
         if (arg.name!==undefined) itemname = arg.name; else
         if (arg.arg!==undefined) itemname = arg.arg; else
         if (arg.item!==undefined) item = arg.item;
      }

      if ((typeof itemname == 'string') && (itemname.indexOf("img:")==0)) {
         // artificial class, can be created by users
         result.obj = {_typename: "TJSImage", fName: itemname.substr(4)};
         return Promise.resolve(result);
      }

      if (item) itemname = this.itemFullName(item);
           else item = this.findItem( { name: itemname, allow_index: true, check_keys: true } );

      // if item not found, try to find nearest parent which could allow us to get inside

      let d = item ? null : this.findItem({ name: itemname, last_exists: true, check_keys: true, allow_index: true });

      // if item not found, try to expand hierarchy central function
      // implements not process get in central method of hierarchy item (if exists)
      // if last_parent found, try to expand it
      if ((d !== null) && ('last' in d) && (d.last !== null)) {
         let parentname = this.itemFullName(d.last);

         // this is indication that expand does not give us better path to searched item
         if ((typeof arg == 'object') && ('rest' in arg))
            if ((arg.rest == d.rest) || (arg.rest.length <= d.rest.length))
               return Promise.resolve(result);

         return this.expandItem(parentname, undefined, options != "hierarchy_expand_verbose").then(res => {
            if (!res) return result;
            let newparentname = this.itemFullName(d.last);
            if (newparentname.length > 0) newparentname += "/";
            return this.getObject( { name: newparentname + d.rest, rest: d.rest }, options);
         });
      }

      result.item = item;

      if ((item !== null) && (typeof item._obj == 'object')) {
         result.obj = item._obj;
         return Promise.resolve(result);
      }

      // normally search _get method in the parent items
      let curr = item;
      while (curr) {
         if (typeof curr._get == 'function')
            return curr._get(item, null, options).then(obj => { result.obj = obj; return result; });
         curr = ('_parent' in curr) ? curr._parent : null;
      }

      return Promise.resolve(result);
   }

      /** @summary returns true if item is last in parent childs list
     * @private */
   isLastSibling(hitem) {
      if (!hitem || !hitem._parent || !hitem._parent._childs) return false;
      let chlds = hitem._parent._childs, indx = chlds.indexOf(hitem);
      if (indx < 0) return false;
      while (++indx < chlds.length)
         if (!('_hidden' in chlds[indx])) return false;
      return true;
   }

   /** @summary Create item html code
     * @private */
   addItemHtml(hitem, d3prnt, arg) {
      if (!hitem || ('_hidden' in hitem)) return true;

      let isroot = (hitem === this.h),
          has_childs = ('_childs' in hitem),
          handle = getDrawHandle(hitem._kind),
          img1 = "", img2 = "", can_click = false, break_list = false,
          d3cont, itemname = this.itemFullName(hitem);

      if (handle) {
         if ('icon' in handle) img1 = handle.icon;
         if ('icon2' in handle) img2 = handle.icon2;
         if ((img1.length==0) && (typeof handle.icon_get == 'function'))
            img1 = handle.icon_get(hitem, this);
         if (canDrawHandle(handle) || ('execute' in handle) || ('aslink' in handle) ||
             (canExpandHandle(handle) && (hitem._more !== false))) can_click = true;
      }

      if ('_icon' in hitem) img1 = hitem._icon;
      if ('_icon2' in hitem) img2 = hitem._icon2;
      if ((img1.length == 0) && ('_online' in hitem))
         hitem._icon = img1 = "img_globe";
      if ((img1.length == 0) && isroot)
         hitem._icon = img1 = "img_base";

      if (hitem._more || hitem._expand || hitem._player || hitem._can_draw)
         can_click = true;

      let can_menu = can_click;
      if (!can_menu && (typeof hitem._kind == 'string') && (hitem._kind.indexOf("ROOT.")==0))
         can_menu = can_click = true;

      if (img2.length == 0) img2 = img1;
      if (img1.length == 0) img1 = (has_childs || hitem._more) ? "img_folder" : "img_page";
      if (img2.length == 0) img2 = (has_childs || hitem._more) ? "img_folderopen" : "img_page";

      if (arg === "update") {
         d3prnt.selectAll("*").remove();
         d3cont = d3prnt;
      } else {
         d3cont = d3prnt.append("div");
         if (arg && (arg >= (hitem._parent._show_limit || settings.HierarchyLimit))) break_list = true;
      }

      hitem._d3cont = d3cont.node(); // set for direct referencing
      d3cont.attr("item", itemname);

      // line with all html elements for this item (excluding childs)
      let d3line = d3cont.append("div").attr('class','h_line');

      // build indent
      let prnt = isroot ? null : hitem._parent;
      while (prnt && (prnt !== this.h)) {
         d3line.insert("div",":first-child")
               .attr("class", this.isLastSibling(prnt) ? "img_empty" : "img_line");
         prnt = prnt._parent;
      }

      let icon_class = "", plusminus = false;

      if (isroot) {
         // for root node no extra code
      } else if (has_childs && !break_list) {
         icon_class = hitem._isopen ? "img_minus" : "img_plus";
         plusminus = true;
      } else /*if (hitem._more) {
         icon_class = "img_plus"; // should be special plus ???
         plusminus = true;
      } else */ {
         icon_class = "img_join";
      }

      let h = this;

      if (icon_class.length > 0) {
         if (break_list || this.isLastSibling(hitem)) icon_class += "bottom";
         let d3icon = d3line.append("div").attr('class', icon_class);
         if (plusminus) d3icon.style('cursor','pointer')
                              .on("click", function(evnt) { h.tree_click(evnt, this, "plusminus"); });
      }

      // make node icons

      if (this.with_icons && !break_list) {
         let icon_name = hitem._isopen ? img2 : img1, d3img;

         if (icon_name.indexOf("img_")==0)
            d3img = d3line.append("div")
                          .attr("class", icon_name)
                          .attr("title", hitem._kind);
         else
            d3img = d3line.append("img")
                          .attr("src", icon_name)
                          .attr("alt","")
                          .attr("title", hitem._kind)
                          .style('vertical-align','top').style('width','18px').style('height','18px');

         if (('_icon_click' in hitem) || (handle && ('icon_click' in handle)))
            d3img.on("click", function(evnt) { h.tree_click(evnt, this, "icon"); });
      }

      let d3a = d3line.append("a");
      if (can_click || has_childs || break_list)
         d3a.attr("class","h_item")
            .on("click", function(evnt) { h.tree_click(evnt, this); });

      if (break_list) {
         hitem._break_point = true; // indicate that list was broken here
         d3a.attr('title', 'there are ' + (hitem._parent._childs.length-arg) + ' more items')
            .text("...more...");
         return false;
      }

      if ('disp_kind' in h) {
         if (settings.DragAndDrop && can_click)
           this.enableDrag(d3a, itemname);

         if (settings.ContextMenu && can_menu)
            d3a.on('contextmenu', function(evnt) { h.tree_contextmenu(evnt, this); });

         d3a.on("mouseover", function() { h.tree_mouseover(true, this); })
            .on("mouseleave", function() { h.tree_mouseover(false, this); });
      } else if (hitem._direct_context && settings.ContextMenu)
         d3a.on('contextmenu', function(evnt) { h.direct_contextmenu(evnt, this); });

      let element_name = hitem._name, element_title = "";

      if ('_realname' in hitem)
         element_name = hitem._realname;

      if ('_title' in hitem)
         element_title = hitem._title;

      if ('_fullname' in hitem)
         element_title += "  fullname: " + hitem._fullname;

      if (!element_title)
         element_title = element_name;

      d3a.attr('title', element_title)
         .text(element_name + ('_value' in hitem ? ":" : ""))
         .style('background', hitem._background ? hitem._background : null);

      if ('_value' in hitem) {
         let d3p = d3line.append("p");
         if ('_vclass' in hitem) d3p.attr('class', hitem._vclass);
         if (!hitem._isopen) d3p.html(hitem._value);
      }

      if (has_childs && (isroot || hitem._isopen)) {
         let d3chlds = d3cont.append("div").attr("class", "h_childs");
         if (this.show_overflow) d3chlds.style("overflow", "initial");
         for (let i = 0; i < hitem._childs.length; ++i) {
            let chld = hitem._childs[i];
            chld._parent = hitem;
            if (!this.addItemHtml(chld, d3chlds, i)) break; // if too many items, skip rest
         }
      }

      return true;
   }

   /** @summary Toggle open state of the item
     * @desc Used with "open all" / "close all" buttons in normal GUI
     * @param {boolean} isopen - if items should be expand or closed
     * @returns {boolean} true when any item was changed */
   toggleOpenState(isopen, h) {
      let hitem = h || this.h;

      if (hitem._childs === undefined) {
         if (!isopen) return false;

         if (this.with_icons) {
            // in normal hierarchy check precisely if item can be expand
            if (!hitem._more && !hitem._expand && !this.canExpandItem(hitem)) return false;
         }

         this.expandItem(this.itemFullName(hitem));
         if (hitem._childs !== undefined) hitem._isopen = true;
         return hitem._isopen;
      }

      if ((hitem !== this.h) && isopen && !hitem._isopen) {
         // when there are childs and they are not see, simply show them
         hitem._isopen = true;
         return true;
      }

      let change_child = false;
      for (let i = 0; i < hitem._childs.length; ++i)
         if (this.toggleOpenState(isopen, hitem._childs[i]))
            change_child = true;

      if ((hitem !== this.h) && !isopen && hitem._isopen && !change_child) {
         // if none of the childs can be closed, than just close that item
         delete hitem._isopen;
         return true;
       }

      if (!h) this.refreshHtml();
      return false;
   }

   /** @summary Refresh HTML code of hierarchy painter
     * @returns {Promise} when done */
   refreshHtml() {

      let d3elem = this.selectDom();
      if (d3elem.empty())
         return Promise.resolve(this);

      d3elem.html("")   // clear html - most simple way
            .style('overflow',this.show_overflow ? 'auto' : 'hidden')
            .style('display','flex')
            .style('flex-direction','column');

      let h = this, factcmds = [], status_item = null;
      this.forEachItem(item => {
         delete item._d3cont; // remove html container
         if (('_fastcmd' in item) && (item._kind == 'Command')) factcmds.push(item);
         if (('_status' in item) && !status_item) status_item = item;
      });

      if (!this.h || d3elem.empty())
         return Promise.resolve(this);

      if (factcmds.length) {
         let fastbtns = d3elem.append("div").attr("style", "display: inline; vertical-align: middle; white-space: nowrap;");
         for (let n = 0; n < factcmds.length; ++n) {
            let btn = fastbtns.append("button")
                       .text("")
                       .attr("class",'jsroot_fastcmd_btn')
                       .attr("item", this.itemFullName(factcmds[n]))
                       .attr("title", factcmds[n]._title)
                       .on("click", function() { h.executeCommand(d3_select(this).attr("item"), this); } );

            if (factcmds[n]._icon)
               btn.style("background-image", `url("${factcmds[n]._icon}")`);
         }
      }

      let d3btns = d3elem.append("p").attr("class", "jsroot").style("margin-bottom","3px").style("margin-top",0);
      d3btns.append("a").attr("class", "h_button").text("open all")
            .attr("title","open all items in the browser").on("click", () => this.toggleOpenState(true));
      d3btns.append("text").text(" | ");
      d3btns.append("a").attr("class", "h_button").text("close all")
            .attr("title","close all items in the browser").on("click", () => this.toggleOpenState(false));

      if (typeof this.removeInspector == 'function') {
         d3btns.append("text").text(" | ");
         d3btns.append("a").attr("class", "h_button").text("remove")
               .attr("title","remove inspector").on("click", () => this.removeInspector());
      }

      if ('_online' in this.h) {
         d3btns.append("text").text(" | ");
         d3btns.append("a").attr("class", "h_button").text("reload")
               .attr("title","reload object list from the server").on("click", () => this.reload());
      }

      if ('disp_kind' in this) {
         d3btns.append("text").text(" | ");
         d3btns.append("a").attr("class", "h_button").text("clear")
               .attr("title","clear all drawn objects").on("click", () => this.clearHierarchy(false));
      }

      let maindiv =
         d3elem.append("div")
               .attr("class", "jsroot")
               .style('font-size', this.with_icons ? "12px" : "15px")
               .style("flex","1");

      if (!this.show_overflow)
         maindiv.style("overflow","auto");

      if (this.background) // case of object inspector and streamer infos display
         maindiv.style("background-color", this.background)
                .style('margin', '2px').style('padding', '2px');

      this.addItemHtml(this.h, maindiv.append("div").attr("class","h_tree"));

      this.setTopPainter(); //assign hpainter as top painter

      if (status_item && !this.status_disabled && !decodeUrl().has('nostatus')) {
         let func = findFunction(status_item._status);
         if (typeof func == 'function')
            return this.createStatusLine().then(sdiv => {
               if (sdiv) func(sdiv, this.itemFullName(status_item));
            });
      }

      return Promise.resolve(this);
   }

   /** @summary Update item node
     * @private */
   updateTreeNode(hitem, d3cont) {
      if ((d3cont === undefined) || d3cont.empty())  {
         d3cont = d3_select(hitem._d3cont ? hitem._d3cont : null);
         let name = this.itemFullName(hitem);
         if (d3cont.empty())
            d3cont = this.selectDom().select("[item='" + name + "']");
         if (d3cont.empty() && ('_cycle' in hitem))
            d3cont = this.selectDom().select("[item='" + name + ";" + hitem._cycle + "']");
         if (d3cont.empty()) return;
      }

      this.addItemHtml(hitem, d3cont, "update");

      if (this.brlayout) this.brlayout.adjustBrowserSize(true);
   }

   /** @summary Update item background
     * @private */
   updateBackground(hitem, scroll_into_view) {
      if (!hitem || !hitem._d3cont) return;

      let d3cont = d3_select(hitem._d3cont);

      if (d3cont.empty()) return;

      let d3a = d3cont.select(".h_item");

      d3a.style('background', hitem._background ? hitem._background : null);

      if (scroll_into_view && hitem._background)
         d3a.node().scrollIntoView(false);
   }

   /** @summary Focus on hierarchy item
     * @param {Object|string} hitem - item to open or its name
     * @desc all parents to the otem will be opened first
     * @returns {Promise} when done
     * @private */
   focusOnItem(hitem) {
      if (typeof hitem == "string")
         hitem = this.findItem(hitem);

      let name = hitem ? this.itemFullName(hitem) : "";
      if (!name) return Promise.resolve(false)

      let itm = hitem, need_refresh = false;

      while (itm) {
         if ((itm._childs !== undefined) && !itm._isopen) {
            itm._isopen = true;
            need_refresh = true;
         }
         itm = itm._parent;
      }

      let promise = need_refresh ? this.refreshHtml() : Promise.resolve(true);

      return promise.then(() => {
         let d3cont = this.selectDom().select("[item='" + name + "']");
         if (d3cont.empty()) return false;
         d3cont.node().scrollIntoView();
         return true;
      });
   }

   /** @summary Handler for click event of item in the hierarchy
     * @private */
   tree_click(evnt, node, place) {
      if (!node) return;

      let d3cont = d3_select(node.parentNode.parentNode),
          itemname = d3cont.attr('item'),
          hitem = itemname ? this.findItem(itemname) : null;
      if (!hitem) return;

      if (hitem._break_point) {
         // special case of more item

         delete hitem._break_point;

         // update item itself
         this.addItemHtml(hitem, d3cont, "update");

         let prnt = hitem._parent, indx = prnt._childs.indexOf(hitem),
             d3chlds = d3_select(d3cont.node().parentNode);

         if (indx < 0) return console.error('internal error');

         prnt._show_limit = (prnt._show_limit || settings.HierarchyLimit) * 2;

         for (let n = indx+1; n < prnt._childs.length; ++n) {
            let chld = prnt._childs[n];
            chld._parent = prnt;
            if (!this.addItemHtml(chld, d3chlds, n)) break; // if too many items, skip rest
         }

         return;
      }

      let prnt = hitem, dflt = undefined;
      while (prnt) {
         if ((dflt = prnt._click_action) !== undefined) break;
         prnt = prnt._parent;
      }

      if (!place || (place=="")) place = "item";
      let selector = (hitem._kind == "ROOT.TKey" && hitem._more) ? "noinspect" : "",
          sett = getDrawSettings(hitem._kind, selector), handle = sett.handle;

      if (place == "icon") {
         let func = null;
         if (typeof hitem._icon_click == 'function')
            func = hitem._icon_click;
         else if (handle && typeof handle.icon_click == 'function')
            func = handle.icon_click;
         if (func && func(hitem,this))
            this.updateTreeNode(hitem, d3cont);
         return;
      }

      // special feature - all items with '_expand' function are not drawn by click
      if ((place=="item") && ('_expand' in hitem) && !evnt.ctrlKey && !evnt.shiftKey) place = "plusminus";

      // special case - one should expand item
      if (((place == "plusminus") && !('_childs' in hitem) && hitem._more) ||
          ((place == "item") && (dflt === "expand"))) {
         return this.expandItem(itemname, d3cont);
      }

      if (place == "item") {

         if ('_player' in hitem)
            return this.player(itemname);

         if (handle && handle.aslink)
            return window.open(itemname + "/");

         if (handle && handle.execute)
            return this.executeCommand(itemname, node.parentNode);

         if (handle && handle.ignore_online && this.isOnlineItem(hitem)) return;

         let can_draw = hitem._can_draw,
             can_expand = hitem._more,
             dflt_expand = (this.default_by_click === "expand"),
             drawopt = "";

         if (evnt.shiftKey) {
            drawopt = (handle && handle.shift) ? handle.shift : "inspect";
            if ((drawopt==="inspect") && handle && handle.noinspect) drawopt = "";
         }
         if (handle && handle.ctrl && evnt.ctrlKey)
            drawopt = handle.ctrl;

         if (!drawopt) {
            for (let pitem = hitem._parent; !!pitem; pitem = pitem._parent) {
               if (pitem._painter) { can_draw = false; if (can_expand===undefined) can_expand = false; break; }
            }
         }

         if (hitem._childs) can_expand = false;

         if (can_draw === undefined) can_draw = sett.draw;
         if (can_expand === undefined) can_expand = sett.expand || sett.get_expand;

         if (can_draw && can_expand && !drawopt) {
            // if default action specified as expand, disable drawing
            // if already displayed, try to expand
            if (dflt_expand || (handle && (handle.dflt === 'expand')) || this.isItemDisplayed(itemname)) can_draw = false;
         }

         if (can_draw && !drawopt && handle && handle.dflt && (handle.dflt !== 'expand'))
            drawopt = handle.dflt;

         if (can_draw)
            return this.display(itemname, drawopt);

         if (can_expand || dflt_expand)
            return this.expandItem(itemname, d3cont);

         // cannot draw, but can inspect ROOT objects
         if ((typeof hitem._kind === "string") && (hitem._kind.indexOf("ROOT.")===0) && sett.inspect && (can_draw !== false))
            return this.display(itemname, "inspect");

         if (!hitem._childs || (hitem === this.h)) return;
      }

      if (hitem._isopen)
         delete hitem._isopen;
      else
         hitem._isopen = true;

      this.updateTreeNode(hitem, d3cont);
   }

   /** @summary Handler for mouse-over event
     * @private */
   tree_mouseover(on, elem) {
      let itemname = d3_select(elem.parentNode.parentNode).attr('item'),
           hitem = this.findItem(itemname);

      if (!hitem) return;

      let painter, prnt = hitem;
      while (prnt && !painter) {
         painter = prnt._painter;
         prnt = prnt._parent;
      }

      if (painter && typeof painter.mouseOverHierarchy === 'function')
         painter.mouseOverHierarchy(on, itemname, hitem);
   }

   /** @summary alternative context menu, used in the object inspector
     * @private */
   direct_contextmenu(evnt, elem) {
      evnt.preventDefault();
      let itemname = d3_select(elem.parentNode.parentNode).attr('item'),
           hitem = this.findItem(itemname);
      if (!hitem) return;

      if (typeof this.fill_context == 'function')
         createMenu(evnt, this).then(menu => {
            this.fill_context(menu, hitem);
            if (menu.size() > 0) {
               menu.tree_node = elem.parentNode;
               menu.show();
            }
         });
   }

   /** @summary Handle context menu in the hieararchy
     * @private */
   tree_contextmenu(evnt, elem) {
      evnt.preventDefault();
      let itemname = d3_select(elem.parentNode.parentNode).attr('item'),
           hitem = this.findItem(itemname);
      if (!hitem) return;

      let onlineprop = this.getOnlineProp(itemname),
          fileprop = this.getFileProp(itemname);

      function qualifyURL(url) {
         const escapeHTML = s => s.split('&').join('&amp;').split('<').join('&lt;').split('"').join('&quot;'),
               el = document.createElement('div');
         el.innerHTML = '<a href="' + escapeHTML(url) + '">x</a>';
         return el.firstChild.href;
      }

      createMenu(evnt, this).then(menu => {

         if ((!itemname || !hitem._parent) && !('_jsonfile' in hitem)) {
            let files = [], addr = "", cnt = 0,
                separ = () => (cnt++ > 0) ? "&" : "?";

            this.forEachRootFile(item => files.push(item._file.fFullURL));

            if (!this.getTopOnlineItem())
               addr = source_dir + "index.htm";

            if (this.isMonitoring())
               addr += separ() + "monitoring=" + this.getMonitoringInterval();

            if (files.length == 1)
               addr += separ() + "file=" + files[0];
            else if (files.length > 1)
               addr += separ() + "files=" + JSON.stringify(files);

            if (this.disp_kind)
               addr += separ() + "layout=" + this.disp_kind.replace(/ /g, "");

            let items = [], opts = [];

            if (this.disp)
               this.disp.forEachFrame(frame => {
                  let dummy = new ObjectPainter(frame),
                      top = dummy.getTopPainter(),
                      item = top ? top.getItemName() : null, opt;

                  if (item) {
                     opt  = top.getDrawOpt() || top.getItemDrawOpt();
                  } else {
                     top = null;
                     dummy.forEachPainter(p => {
                        let _item = p.getItemName();
                        if (!_item) return;
                        let _opt = p.getDrawOpt() || p.getItemDrawOpt() || "";
                        if (!top) {
                           top = p;
                           item = _item;
                           opt = _opt;
                        } else if (top.getPadPainter() === p.getPadPainter()) {
                           if (_opt.indexOf("same ")==0) _opt = _opt.substr(5);
                           item += "+" + _item;
                           opt += "+" + _opt;
                        }
                     });
                  }

                  if (item) {
                     items.push(item);
                     opts.push(opt || "");
                  }
               });

            if (items.length == 1) {
               addr += separ() + "item=" + items[0] + separ() + "opt=" + opts[0];
            } else if (items.length > 1) {
               addr += separ() + "items=" + JSON.stringify(items) + separ() + "opts=" + JSON.stringify(opts);
            }

            menu.add("Direct link", () => window.open(addr));
            menu.add("Only items", () => window.open(addr + "&nobrowser"));
         } else if (onlineprop) {
            this.fillOnlineMenu(menu, onlineprop, itemname);
         } else {
            let sett = getDrawSettings(hitem._kind, 'nosame');

            // allow to draw item even if draw function is not defined
            if (hitem._can_draw) {
               if (!sett.opts) sett.opts = [""];
               if (sett.opts.indexOf("") < 0) sett.opts.unshift("");
            }

            if (sett.opts)
               menu.addDrawMenu("Draw", sett.opts, arg => this.display(itemname, arg));

            if (fileprop && sett.opts && !fileprop.localfile) {
               let filepath = qualifyURL(fileprop.fileurl);
               if (filepath.indexOf(source_dir) == 0)
                  filepath = filepath.slice(source_dir.length);
               filepath = fileprop.kind + "=" + filepath;
               if (fileprop.itemname.length > 0) {
                  let name = fileprop.itemname;
                  if (name.search(/\+| |\,/)>=0) name = "\'" + name + "\'";
                  filepath += "&item=" + name;
               }

               menu.addDrawMenu("Draw in new tab", sett.opts,
                                arg => window.open(source_dir + "index.htm?nobrowser&"+filepath +"&opt="+arg));
            }

            if ((sett.expand || sett.get_expand) && !('_childs' in hitem) && (hitem._more || !('_more' in hitem)))
               menu.add("Expand", () => this.expandItem(itemname));

            if (hitem._kind === "ROOT.TStyle")
               menu.add("Apply", () => this.applyStyle(itemname));
         }

         if (typeof hitem._menu == 'function')
            hitem._menu(menu, hitem, this);

         if (menu.size() > 0) {
            menu.tree_node = elem.parentNode;
            if (menu.separ) menu.add("separator"); // add separator at the end
            menu.add("Close");
            menu.show();
         }

      }); // end menu creation

      return false;
   }

   /** @summary Starts player for specified item
     * @desc Same as "Player" context menu
     * @param {string} itemname - item name for which player should be started
     * @param {string} [option] - extra options for the player
     * @returns {Promise} when ready*/
   async player(itemname, option) {
      let item = this.findItem(itemname);

      if (!item || !item._player || (typeof item._player != 'string'))
         return null;

      let player_func = findFunction(item._player);

      if (!player_func)
         if (item._prereq || (item._player.indexOf('JSROOT.') >= 0)) {
            await this.loadScripts("", item._prereq);
            player_func = findFunction(item._player);
         }
      if (!player_func && item._module) {
         let hh = import(item._module);
         player_func = hh ? hh[item._player] : null;
      }
      if (typeof player_func != 'function')
         return null;

      let mdi = await this.createDisplay();
      return mdi ? player_func(this, itemname, option) : null;
   }

   /** @summary Checks if item can be displayed with given draw option
     * @private */
   canDisplay(item, drawopt) {
      if (!item) return false;
      if (item._player) return true;
      if (item._can_draw !== undefined) return item._can_draw;
      if (drawopt == 'inspect') return true;
      const handle = getDrawHandle(item._kind, drawopt);
      return canDrawHandle(handle);
   }

   /** @summary Returns true if given item displayed
     * @param {string} itemname - item name */
   isItemDisplayed(itemname) {
      let mdi = this.getDisplay();
      return mdi ? mdi.findFrame(itemname) !== null : false;
   }

   /** @summary Display specified item
     * @param {string} itemname - item name
     * @param {string} [drawopt] - draw option for the item
     * @returns {Promise} with created painter object */
   display(itemname, drawopt) {
      let h = this,
          painter = null,
          updating = false,
          item = null,
          display_itemname = itemname,
          frame_name = itemname,
          marker = "::_display_on_frame_::",
          p = drawopt ? drawopt.indexOf(marker) : -1;

      if (p >= 0) {
         frame_name = drawopt.substr(p + marker.length);
         drawopt = drawopt.substr(0, p);
      }

      function complete(respainter, err) {
         if (err) console.log('When display ', itemname, 'got', err);

         if (updating && item) delete item._doing_update;
         if (!updating) showProgress();
         if (respainter && (typeof respainter === 'object') && (typeof respainter.setItemName === 'function')) {
            respainter.setItemName(display_itemname, updating ? null : drawopt, h); // mark painter as created from hierarchy
            if (item && !item._painter) item._painter = respainter;
         }

         return respainter || painter;
      }

      return h.createDisplay().then(mdi => {

         if (!mdi) return complete();

         item = h.findItem(display_itemname);

         if (item && ('_player' in item))
            return h.player(display_itemname, drawopt).then(res => complete(res));

         updating = (typeof drawopt == 'string') && (drawopt.indexOf("update:")==0);

         if (updating) {
            drawopt = drawopt.substr(7);
            if (!item || item._doing_update) return complete();
            item._doing_update = true;
         }

         if (item && !h.canDisplay(item, drawopt)) return complete();

         let divid = "", use_dflt_opt = false;
         if ((typeof drawopt == 'string') && (drawopt.indexOf("divid:") >= 0)) {
            let pos = drawopt.indexOf("divid:");
            divid = drawopt.slice(pos+6);
            drawopt = drawopt.slice(0, pos);
         }

         if (drawopt == "__default_draw_option__") {
            use_dflt_opt = true;
            drawopt = "";
         }

         if (!updating) showProgress("Loading " + display_itemname);

         return h.getObject(display_itemname, drawopt).then(result => {

            if (!updating) showProgress();

            if (!item) item = result.item;
            let obj = result.obj;

            if (!obj) return complete();

            if (!updating) showProgress("Drawing " + display_itemname);

            let handle = obj._typename ? getDrawHandle("ROOT." + obj._typename) : null;

            if (handle && handle.draw_field && obj[handle.draw_field]) {
               obj = obj[handle.draw_field];
               if (!drawopt) drawopt = handle.draw_field_opt || "";
               handle = obj._typename ? getDrawHandle("ROOT." + obj._typename) : null;
            }

            if (use_dflt_opt && handle && handle.dflt && !drawopt && (handle.dflt != 'expand'))
               drawopt = handle.dflt;

            if (divid.length > 0) {
               let func = updating ? redraw : draw;
               return func(divid, obj, drawopt).then(p => complete(p)).catch(err => complete(null, err));
            }

            mdi.forEachPainter((p, frame) => {
               if (p.getItemName() != display_itemname) return;
               // verify that object was drawn with same option as specified now (if any)
               if (!updating && drawopt && (p.getItemDrawOpt() != drawopt)) return;

               // do not actiavte frame when doing update
               // mdi.activateFrame(frame);

               if ((typeof p.redrawObject == 'function') && p.redrawObject(obj, drawopt)) painter = p;
            });

            if (painter) return complete();

            if (updating) {
               console.warn(`something went wrong - did not found painter when doing update of ${display_itemname}`);
               return complete();
            }

            let frame = mdi.findFrame(frame_name, true);
            d3_select(frame).html("");
            mdi.activateFrame(frame);

            return draw(frame, obj, drawopt)
                         .then(p => complete(p))
                         .catch(err => complete(null, err));

         });
      });
   }

   /** @summary Enable drag of the element
     * @private  */
   enableDrag(d3elem /*, itemname*/) {
      d3elem.attr("draggable", "true").on("dragstart", function(ev) {
         let itemname = this.parentNode.parentNode.getAttribute('item');
         ev.dataTransfer.setData("item", itemname);
      });
   }

   /** @summary Enable drop on the frame
     * @private  */
   enableDrop(frame) {
      let h = this;
      d3_select(frame).on("dragover", function(ev) {
         let itemname = ev.dataTransfer.getData("item"),
              ditem = h.findItem(itemname);
         if (ditem && (typeof ditem._kind == 'string') && (ditem._kind.indexOf("ROOT.")==0))
            ev.preventDefault(); // let accept drop, otherwise it will be refuced
      }).on("dragenter", function() {
         d3_select(this).classed('jsroot_drag_area', true);
      }).on("dragleave", function() {
         d3_select(this).classed('jsroot_drag_area', false);
      }).on("drop", function(ev) {
         d3_select(this).classed('jsroot_drag_area', false);
         let itemname = ev.dataTransfer.getData("item");
         if (itemname) h.dropItem(itemname, this.getAttribute("id"));
      });
   }

   /** @summary Remove all drop handlers on the frame
     * @private  */
   clearDrop(frame) {
      d3_select(frame).on("dragover", null).on("dragenter", null).on("dragleave", null).on("drop", null);
   }

  /** @summary Drop item on specified element for drawing
    * @returns {Promise} when completed
    * @private */
   dropItem(itemname, divid, opt) {

      if (opt && typeof opt === 'function') { call_back = opt; opt = ""; }
      if (opt===undefined) opt = "";

      let drop_complete = (drop_painter, is_main_painter) => {
         if (drop_painter && !is_main_painter && (typeof drop_painter === 'object') && (typeof drop_painter.setItemName == 'function'))
            drop_painter.setItemName(itemname, null, this);
         return drop_painter;
      }

      if (itemname == "$legend")
         return produceLegend(divid, opt).then(legend_painter => drop_complete(legend_painter));

      return this.getObject(itemname).then(res => {

         if (!res.obj) return null;

         let main_painter = getElementMainPainter(divid);

         if (main_painter && (typeof main_painter.performDrop === 'function'))
            return main_painter.performDrop(res.obj, itemname, res.item, opt).then(p => drop_complete(p, main_painter === p));

         if (main_painter && main_painter.accept_drops)
            return draw(divid, res.obj, "same " + opt).then(p => drop_complete(p, main_painter === p));

         this.cleanupFrame(divid);
         return draw(divid, res.obj, opt).then(p => drop_complete(p));
      });
   }

   /** @summary Update specified items
     * @desc Method can be used to fetch new objects and update all existing drawings
     * @param {string|array|boolean} arg - either item name or array of items names to update or true if only automatic items will be updated
     * @returns {Promise} when ready */
   updateItems(arg) {

      if (!this.disp)
         return Promise.resolve(false);

      let allitems = [], options = [], only_auto_items = false, want_update_all = false;

      if (typeof arg == "string")
         arg = [ arg ];
      else if (typeof arg != 'object') {
         if (arg === undefined) arg = !this.isMonitoring();
         want_update_all = true;
         only_auto_items = !!arg;
      }

      // first collect items
      this.disp.forEachPainter(p => {
         let itemname = p.getItemName();

         if ((typeof itemname != 'string') || (allitems.indexOf(itemname) >= 0)) return;

         if (want_update_all) {
            let item = this.findItem(itemname);
            if (!item || ('_not_monitor' in item) || ('_player' in item)) return;
            if (!('_always_monitor' in item)) {
               let forced = false, handle = getDrawHandle(item._kind);
               if (handle && ('monitor' in handle)) {
                  if ((handle.monitor === false) || (handle.monitor == 'never')) return;
                  if (handle.monitor == 'always') forced = true;
               }
               if (!forced && only_auto_items) return;
            }
         } else {
            if (arg.indexOf(itemname) < 0) return;
         }

         allitems.push(itemname);
         options.push("update:" + p.getItemDrawOpt());
      }, true); // only visible panels are considered

      // force all files to read again (normally in non-browser mode)
      if (this.files_monitoring && !only_auto_items && want_update_all)
         this.forEachRootFile(item => {
            this.forEachItem(fitem => { delete fitem._readobj; }, item);
            delete item._file;
         });

      return this.displayItems(allitems, options);
   }

   /** @summary Display all provided elements
     * @returns {Promise} when drawing finished
     * @private */
   displayItems(items, options) {

      if (!items || (items.length == 0))
         return Promise.resolve(true);

      let h = this;

      if (!options) options = [];
      while (options.length < items.length)
         options.push("__default_draw_option__");

      if ((options.length == 1) && (options[0] == "iotest")) {
         this.clearHierarchy();
         d3_select("#" + this.disp_frameid).html("<h2>Start I/O test</h2>");

         let tm0 = new Date();
         return this.getObject(items[0]).then(() => {
            let tm1 = new Date();
            d3_select("#" + this.disp_frameid).append("h2").html("Item " + items[0] + " reading time = " + (tm1.getTime() - tm0.getTime()) + "ms");
            return Promise.resolve(true);
         });
      }

      let dropitems = new Array(items.length),
          dropopts = new Array(items.length),
          images = new Array(items.length);

      // First of all check that items are exists, look for cycle extension and plus sign
      for (let i = 0; i < items.length; ++i) {
         dropitems[i] = dropopts[i] = null;

         let item = items[i], can_split = true;

         if (item && item.indexOf("img:")==0) { images[i] = true; continue; }

         if (item && (item.length>1) && (item[0]=='\'') && (item[item.length-1]=='\'')) {
            items[i] = item.substr(1, item.length-2);
            can_split = false;
         }

         let elem = h.findItem({ name: items[i], check_keys: true });
         if (elem) { items[i] = h.itemFullName(elem); continue; }

         if (can_split && (items[i][0]=='[') && (items[i][items[i].length-1]==']')) {
            dropitems[i] = parseAsArray(items[i]);
            items[i] = dropitems[i].shift();
         } else if (can_split && (items[i].indexOf("+") > 0)) {
            dropitems[i] = items[i].split("+");
            items[i] = dropitems[i].shift();
         }

         if (dropitems[i] && dropitems[i].length > 0) {
            // allow to specify _same_ item in different file
            for (let j = 0; j < dropitems[i].length; ++j) {
               let pos = dropitems[i][j].indexOf("_same_");
               if ((pos > 0) && (h.findItem(dropitems[i][j]) === null))
                  dropitems[i][j] = dropitems[i][j].substr(0,pos) + items[i].substr(pos);

               elem = h.findItem({ name: dropitems[i][j], check_keys: true });
               if (elem) dropitems[i][j] = h.itemFullName(elem);
            }

            if ((options[i][0] == "[") && (options[i][options[i].length-1] == "]")) {
               dropopts[i] = parseAsArray(options[i]);
               options[i] = dropopts[i].shift();
            } else if (options[i].indexOf("+") > 0) {
               dropopts[i] = options[i].split("+");
               options[i] = dropopts[i].shift();
            } else {
               dropopts[i] = [];
            }

            while (dropopts[i].length < dropitems[i].length) dropopts[i].push("");
         }

         // also check if subsequent items has _same_, than use name from first item
         let pos = items[i].indexOf("_same_");
         if ((pos > 0) && !h.findItem(items[i]) && (i > 0))
            items[i] = items[i].substr(0,pos) + items[0].substr(pos);

         elem = h.findItem({ name: items[i], check_keys: true });
         if (elem) items[i] = h.itemFullName(elem);
      }

      // now check that items can be displayed
      for (let n = items.length - 1; n >= 0; --n) {
         if (images[n]) continue;
         let hitem = h.findItem(items[n]);
         if (!hitem || h.canDisplay(hitem, options[n])) continue;
         // try to expand specified item
         h.expandItem(items[n], null, true);
         items.splice(n, 1);
         options.splice(n, 1);
         dropitems.splice(n, 1);
      }

      if (items.length == 0)
         return Promise.resolve(true);

      let frame_names = new Array(items.length), items_wait = new Array(items.length);
      for (let n = 0; n < items.length; ++n) {
         items_wait[n] = 0;
         let fname = items[n], k = 0;
         if (items.indexOf(fname) < n) items_wait[n] = true; // if same item specified, one should wait first drawing before start next
         let p = options[n].indexOf("frameid:");
         if (p >= 0) {
            fname = options[n].substr(p+8);
            options[n] = options[n].substr(0,p);
         } else {
            while (frame_names.indexOf(fname)>=0)
               fname = items[n] + "_" + k++;
         }
         frame_names[n] = fname;
      }

      // now check if several same items present - select only one for the drawing
      // if draw option includes 'main', such item will be drawn first
      for (let n = 0; n < items.length; ++n) {
         if (items_wait[n] !== 0) continue;
         let found_main = n;
         for (let k = 0; k < items.length; ++k)
            if ((items[n]===items[k]) && (options[k].indexOf('main')>=0)) found_main = k;
         for (let k = 0; k < items.length; ++k)
            if (items[n]===items[k]) items_wait[k] = (found_main != k);
      }

      return this.createDisplay().then(mdi => {
         if (!mdi) return false;

         // Than create empty frames for each item
         for (let i = 0; i < items.length; ++i)
            if (options[i].indexOf('update:')!==0) {
               mdi.createFrame(frame_names[i]);
               options[i] += "::_display_on_frame_::"+frame_names[i];
            }

         function DropNextItem(indx, painter) {
            if (painter && dropitems[indx] && (dropitems[indx].length > 0))
               return h.dropItem(dropitems[indx].shift(), painter.getDom(), dropopts[indx].shift()).then(() => DropNextItem(indx, painter));

            dropitems[indx] = null; // mark that all drop items are processed
            items[indx] = null; // mark item as ready

            for (let cnt = 0; cnt < items.length; ++cnt) {
               if (items[cnt]===null) continue; // ignore completed item
               if (items_wait[cnt] && items.indexOf(items[cnt])===cnt) {
                  items_wait[cnt] = false;
                  return h.display(items[cnt], options[cnt]).then(painter => DropNextItem(cnt, painter));
               }
            }
         }

         let promises = [];

         // We start display of all items parallel, but only if they are not the same
         for (let i = 0; i < items.length; ++i)
            if (!items_wait[i])
               promises.push(h.display(items[i], options[i]).then(painter => DropNextItem(i, painter)));

         return Promise.all(promises);
      });
   }

   /** @summary Reload hierarchy and refresh html code
     * @returns {Promise} when completed */
   reload() {
      if ('_online' in this.h)
         return this.openOnline(this.h._online).then(() => this.refreshHtml());
      return Promise.resolve(false);
   }

   /** @summary activate (select) specified item
     * @param {Array} items - array of items names
     * @param {boolean} [force] - if specified, all required sub-levels will be opened
     * @private */
   activateItems(items, force) {

      if (typeof items == 'string') items = [ items ];

      let active = [], // array of elements to activate
          update = []; // array of elements to update
      this.forEachItem(item => { if (item._background) { active.push(item); delete item._background; } });

      let mark_active = () => {

         for (let n = update.length-1; n >= 0; --n)
            this.updateTreeNode(update[n]);

         for (let n = 0; n < active.length; ++n)
            this.updateBackground(active[n], force);
      }

      let find_next = (itemname, prev_found) => {
         if (itemname === undefined) {
            // extract next element
            if (items.length == 0) return mark_active();
            itemname = items.shift();
         }

         let hitem = this.findItem(itemname);

         if (!hitem) {
            let d = this.findItem({ name: itemname, last_exists: true, check_keys: true, allow_index: true });
            if (!d || !d.last) return find_next();
            d.now_found = this.itemFullName(d.last);

            if (force) {

               // if after last expand no better solution found - skip it
               if ((prev_found!==undefined) && (d.now_found === prev_found)) return find_next();

               return this.expandItem(d.now_found).then(res => {
                  if (!res) return find_next();
                  let newname = this.itemFullName(d.last);
                  if (newname.length > 0) newname+="/";
                  find_next(newname + d.rest, d.now_found);
               });
            }
            hitem = d.last;
         }

         if (hitem) {
            // check that item is visible (opened), otherwise should enable parent

            let prnt = hitem._parent;
            while (prnt) {
               if (!prnt._isopen) {
                  if (force) {
                     prnt._isopen = true;
                     if (update.indexOf(prnt)<0) update.push(prnt);
                  } else {
                     hitem = prnt; break;
                  }
               }
               prnt = prnt._parent;
            }

            hitem._background = 'LightSteelBlue';
            if (active.indexOf(hitem)<0) active.push(hitem);
         }

         find_next();
      }

      if (force && this.brlayout) {
         if (!this.brlayout.browser_kind)
           return this.createBrowser('float', true).then(() => find_next());
         if (!this.brlayout.browser_visible)
            this.brlayout.toggleBrowserVisisbility();
      }

      // use recursion
      find_next();
   }

   /** @summary Check if item can be (potentially) expand
     * @private */
   canExpandItem(item) {
      if (!item) return false;
      if (item._expand) return true;
      let handle = getDrawHandle(item._kind, "::expand");
      return handle && canExpandHandle(handle);
   }

   /** @summary expand specified item
     * @param {String} itemname - item name
     * @returns {Promise} when ready */
   expandItem(itemname, d3cont, silent) {
      let hitem = this.findItem(itemname), hpainter = this;

      if (!hitem && d3cont)
         return Promise.resolve();

      async function DoExpandItem(_item, _obj){

         if (typeof _item._expand == 'string')
            _item._expand = findFunction(item._expand);

         if (typeof _item._expand !== 'function') {
            let handle = getDrawHandle(_item._kind, "::expand");

            if (handle && handle.expand_item) {
               _obj = _obj[handle.expand_item];
              handle = (_obj && _obj._typename) ? getDrawHandle("ROOT."+_obj._typename, "::expand") : null;
            }

            if (handle && (handle.expand || handle.get_expand)) {
               if (typeof handle.expand == 'function')
                  _item._expand = handle.expand;
               else if (typeof handle.expand == 'string') {
                  let v6 = await _ensureJSROOT();
                  await v6.require(handle.prereq);
                  await v6._complete_loading();
                  _item._expand = handle.expand = findFunction(handle.expand);
               } else if (typeof handle.get_expand == 'function') {
                  _item._expand = handle.expand = await handle.get_expand();
               }
            }
         }

         // try to use expand function
         if (_obj && _item && (typeof _item._expand === 'function')) {
            if (_item._expand(_item, _obj)) {
               _item._isopen = true;
               if (_item._parent && !_item._parent._isopen) {
                  _item._parent._isopen = true; // also show parent
                  if (!silent) hpainter.updateTreeNode(_item._parent);
               } else {
                  if (!silent) hpainter.updateTreeNode(_item, d3cont);
               }
               return _item;
            }
         }

         if (_obj && objectHierarchy(_item, _obj)) {
            _item._isopen = true;
            if (_item._parent && !_item._parent._isopen) {
               _item._parent._isopen = true; // also show parent
               if (!silent) hpainter.updateTreeNode(_item._parent);
            } else {
               if (!silent) hpainter.updateTreeNode(_item, d3cont);
            }
            return _item;
         }

         return -1;
      };

      let promise = Promise.resolve(-1);

      if (hitem) {
         // item marked as it cannot be expanded, also top item cannot be changed
         if ((hitem._more === false) || (!hitem._parent && hitem._childs))
            return Promise.resolve();

         if (hitem._childs && hitem._isopen) {
            hitem._isopen = false;
            if (!silent) this.updateTreeNode(hitem, d3cont);
            return Promise.resolve();
         }

         if (hitem._obj) promise = DoExpandItem(hitem, hitem._obj);
      }

      return promise.then(res => {
         if (res !== -1) return res; // done

         showProgress("Loading " + itemname);

         return this.getObject(itemname, silent ? "hierarchy_expand" : "hierarchy_expand_verbose").then(res => {

            showProgress();

            if (res.obj) return DoExpandItem(res.item, res.obj).then(res => { return (res !== -1) ? res : undefined; });
         });
      });

   }

   /** @summary Return main online item
     * @private */
   getTopOnlineItem(item) {
      if (item) {
         while (item && (!('_online' in item))) item = item._parent;
         return item;
      }

      if (!this.h) return null;
      if ('_online' in this.h) return this.h;
      if (this.h._childs && ('_online' in this.h._childs[0])) return this.h._childs[0];
      return null;
   }

   /** @summary Call function for each item which corresponds to JSON file
     * @private */
   forEachJsonFile(func) {
      if (!this.h) return;
      if ('_jsonfile' in this.h)
         return func(this.h);

      if (this.h._childs)
         for (let n = 0; n < this.h._childs.length; ++n) {
            let item = this.h._childs[n];
            if ('_jsonfile' in item) func(item);
         }
   }

   /** @summary Open JSON file
     * @param {string} filepath - URL to JSON file
     * @returns {Promise} when object ready */
   openJsonFile(filepath) {
      let isfileopened = false;
      this.forEachJsonFile(item => { if (item._jsonfile==filepath) isfileopened = true; });
      if (isfileopened) return Promise.resolve();

      return httpRequest(filepath, 'object').then(res => {
         let h1 = { _jsonfile: filepath, _kind: "ROOT." + res._typename, _jsontmp: res, _name: filepath.split("/").pop() };
         if (res.fTitle) h1._title = res.fTitle;
         h1._get = function(item /* ,itemname */) {
            if (item._jsontmp)
               return Promise.resolve(item._jsontmp);
            return httpRequest(item._jsonfile, 'object')
                         .then(res => {
                             item._jsontmp = res;
                             return res;
                          });
         }
         if (!this.h) this.h = h1; else
         if (this.h._kind == 'TopFolder') this.h._childs.push(h1); else {
            let h0 = this.h, topname = ('_jsonfile' in h0) ? "Files" : "Items";
            this.h = { _name: topname, _kind: 'TopFolder', _childs : [h0, h1] };
         }

         return this.refreshHtml();
      });
   }

   /** @summary Call function for each item which corresponds to ROOT file
     * @private */
   forEachRootFile(func) {
      if (!this.h) return;
      if ((this.h._kind == "ROOT.TFile") && this.h._file)
         return func(this.h);

      if (this.h._childs)
         for (let n = 0; n < this.h._childs.length; ++n) {
            let item = this.h._childs[n];
            if ((item._kind == 'ROOT.TFile') && ('_fullurl' in item))
               func(item);
         }
   }

   /** @summary Open ROOT file
     * @param {string} filepath - URL to ROOT file
     * @returns {Promise} when file is opened */
   openRootFile(filepath) {

      let isfileopened = false;
      this.forEachRootFile(item => { if (item._fullurl===filepath) isfileopened = true; });
      if (isfileopened) return Promise.resolve();

      showProgress("Opening " + filepath + " ...");

      return openFile(filepath).then(file => {

         let h1 = this.fileHierarchy(file);
         h1._isopen = true;
         if (!this.h) {
            this.h = h1;
            if (this._topname) h1._name = this._topname;
         } else if (this.h._kind == 'TopFolder') {
            this.h._childs.push(h1);
         }  else {
            let h0 = this.h, topname = (h0._kind == "ROOT.TFile") ? "Files" : "Items";
            this.h = { _name: topname, _kind: 'TopFolder', _childs : [h0, h1], _isopen: true };
         }

         return this.refreshHtml();
      }).catch(() => {
         // make CORS warning
         if (isBatchMode())
            console.error(`Fail to open ${filepath} - check CORS headers`);
         else if (!d3_select("#gui_fileCORS").style("background","red").empty())
            setTimeout(() => d3_select("#gui_fileCORS").style("background",''), 5000);
         return false;
      }).finally(() => showProgress());
   }

   /** @summary Apply loaded TStyle object
     * @desc One also can specify item name of JSON file name where style is loaded
     * @param {object|string} style - either TStyle object of item name where object can be load */
   async applyStyle(style) {
      if (!style)
         return;

      if (typeof style === 'string') {
         let item = this.findItem({ name: style, allow_index: true, check_keys: true });
         if (item !== null)
            style = await this.getObject(item).then(res => res.obj);
         else if (style.indexOf('.json') > 0)
            style = await httpRequest(style, 'object');
      }

      if (style && (typeof style === 'object') && (style._typename === "TStyle"))
         Object.assign(gStyle, style);

   }

   /** @summary Provides information abouf file item
     * @private */
   getFileProp(itemname) {
      let item = this.findItem(itemname);
      if (!item) return null;

      let subname = item._name;
      while (item._parent) {
         item = item._parent;
         if ('_file' in item)
            return { kind: "file", fileurl: item._file.fURL, itemname: subname, localfile: !!item._file.fLocalFile };

         if ('_jsonfile' in item)
            return { kind: "json", fileurl: item._jsonfile, itemname: subname };

         subname = item._name + "/" + subname;
      }

      return null;
   }

   /** @summary Provides URL for online item
     * @desc Such URL can be used  to request data from the server
     * @returns string or null if item is not online
     * @private */
   getOnlineItemUrl(item) {
      if (typeof item == "string") item = this.findItem(item);
      let prnt = item;
      while (prnt && (prnt._online===undefined)) prnt = prnt._parent;
      return prnt ? (prnt._online + this.itemFullName(item, prnt)) : null;
   }

   /** @summary Returns true if item is online
     * @private */
   isOnlineItem(item) {
      return this.getOnlineItemUrl(item) !== null;
   }

   /** @summary method used to request object from the http server
     * @returns {Promise} with requested object
     * @private */
   getOnlineItem(item, itemname, option) {

      let url = itemname, h_get = false, req = "", req_kind = "object", draw_handle = null;

      if ((typeof option == "string") && (option.indexOf('hierarchy_expand')==0)) {
         h_get = true;
         option = undefined;
      }

      if (item) {
         url = this.getOnlineItemUrl(item);
         let func = null;
         if ('_kind' in item) draw_handle = getDrawHandle(item._kind);

         if (h_get) {
            req = 'h.json?compact=3';
            item._expand = onlineHierarchy; // use proper expand function
         } else if (item._make_request) {
            func = findFunction(item._make_request);
         } else if (draw_handle && draw_handle.make_request) {
            func = draw_handle.make_request;
         }

         if (typeof func == 'function') {
            // ask to make request
            let dreq = func(this, item, url, option);
            // result can be simple string or object with req and kind fields
            if (dreq!=null)
               if (typeof dreq == 'string') req = dreq; else {
                  if ('req' in dreq) req = dreq.req;
                  if ('kind' in dreq) req_kind = dreq.kind;
               }
         }

         if ((req.length == 0) && (item._kind.indexOf("ROOT.") != 0))
           req = 'item.json.gz?compact=3';
      }

      if (!itemname && item && ('_cached_draw_object' in this) && (req.length == 0)) {
         // special handling for online draw when cashed
         let obj = this._cached_draw_object;
         delete this._cached_draw_object;
         return Promise.resolve(obj);
      }

      if (req.length == 0) req = 'root.json.gz?compact=23';

      if (url.length > 0) url += "/";
      url += req;

      return new Promise(resolveFunc => {

         let itemreq = createHttpRequest(url, req_kind, obj => {

            let func = null;

            if (!h_get && item && ('_after_request' in item)) {
               func = findFunction(item._after_request);
            } else if (draw_handle && ('after_request' in draw_handle))
               func = draw_handle.after_request;

            if (typeof func == 'function') {
               let res = func(this, item, obj, option, itemreq);
               if (res && (typeof res == "object")) obj = res;
            }

            resolveFunc(obj);
         });

         itemreq.send(null);
      });
   }

   /** @summary Access THttpServer with provided address
     * @param {string} server_address - URL to server like "http://localhost:8090/"
     * @returns {Promise} when ready */
   openOnline(server_address) {
      let AdoptHierarchy = result => {
         this.h = result;
         if (!result) return Promise.resolve(null);

         if (('_title' in this.h) && (this.h._title!='')) document.title = this.h._title;

         result._isopen = true;

         // mark top hierarchy as online data and
         this.h._online = server_address;

         this.h._get = (item, itemname, option) => {
            return this.getOnlineItem(item, itemname, option);
         };

         this.h._expand = onlineHierarchy;

         let styles = [], scripts = [], v6_modules = [];
         this.forEachItem(item => {
            if (item._childs !== undefined)
               item._expand = onlineHierarchy;

            if (item._autoload) {
               let arr = item._autoload.split(";");
               arr.forEach(name => {
                  if ((name.length > 3) && (name.lastIndexOf(".js") == name.length-3)) {
                     if (!scripts.find(elem => elem == name)) scripts.push(name);
                  } else if ((name.length > 4) && (name.lastIndexOf(".css") == name.length-4)) {
                     if (!styles.find(elem => elem == name)) styles.push(name);
                  } else if (name && !v6_modules.find(elem => elem == name)) {
                     v6_modules.push(name);
                  }
               });
            }
         });

         return this.loadScripts(scripts, v6_modules)
               .then(() => loadScript(styles))
               .then(() => {
                  this.forEachItem(item => {
                     if (!('_drawfunc' in item) || !('_kind' in item)) return;
                     let typename = "kind:" + item._kind;
                     if (item._kind.indexOf('ROOT.')==0) typename = item._kind.slice(5);
                     let drawopt = item._drawopt;
                     if (!canDrawHandle(typename) || drawopt)
                        addDrawFunc({ name: typename, func: item._drawfunc, script: item._drawscript, opt: drawopt });
                  });

                  return this;
               });
      }

      if (!server_address) server_address = "";

      if (typeof server_address == 'object') {
         let h = server_address;
         server_address = "";
         return AdoptHierarchy(h);
      }

      return httpRequest(server_address + "h.json?compact=3", 'object').then(hh => AdoptHierarchy(hh));
   }

   /** @summary Get properties for online item  - server name and relative name
     * @private */
   getOnlineProp(itemname) {
      let item = this.findItem(itemname);
      if (!item) return null;

      let subname = item._name;
      while (item._parent) {
         item = item._parent;

         if ('_online' in item) {
            return {
               server: item._online,
               itemname: subname
            };
         }
         subname = item._name + "/" + subname;
      }

      return null;
   }

   /** @summary Fill context menu for online item
     * @private */
   fillOnlineMenu(menu, onlineprop, itemname) {

      let node = this.findItem(itemname),
          sett = getDrawSettings(node._kind, 'nosame;noinspect'),
          handle = getDrawHandle(node._kind),
          root_type = (typeof node._kind == 'string') ? node._kind.indexOf("ROOT.") == 0 : false;

      if (sett.opts && (node._can_draw !== false)) {
         sett.opts.push('inspect');
         menu.addDrawMenu("Draw", sett.opts, arg => this.display(itemname, arg));
      }

      if (!node._childs && (node._more !== false) && (node._more || root_type || sett.expand || sett.get_expand))
         menu.add("Expand", () => this.expandItem(itemname));

      if (handle && ('execute' in handle))
         menu.add("Execute", () => this.executeCommand(itemname, menu.tree_node));

      if (sett.opts && (node._can_draw !== false))
         menu.addDrawMenu("Draw in new window", sett.opts,
                           arg => window.open(onlineprop.server + "?nobrowser&item=" + onlineprop.itemname +
                                              (this.isMonitoring() ? "&monitoring=" + this.getMonitoringInterval() : "") +
                                              (arg ? "&opt=" + arg : "")));

      if (sett.opts && (sett.opts.length > 0) && root_type && (node._can_draw !== false))
         menu.addDrawMenu("Draw as png", sett.opts,
                           arg => window.open(onlineprop.server + onlineprop.itemname + "/root.png?w=600&h=400" + (arg ? "&opt=" + arg : "")));

      if ('_player' in node)
         menu.add("Player", () => this.player(itemname));
   }

   /** @summary Assign existing hierarchy to the painter and refresh HTML code
     * @private */
   setHierarchy(h) {
      this.h = h;
      this.refreshHtml();
   }

   /** @summary Configures monitoring interval
     * @param {number} interval - repetition interval in ms
     * @param {boolean} flag - initial monitoring state */
   setMonitoring(interval, monitor_on) {

      this._runMonitoring("cleanup");

      if (interval) {
         interval = parseInt(interval);
         if (Number.isInteger(interval) && (interval > 0)) {
            this._monitoring_interval = Math.max(100,interval);
            monitor_on = true;
         } else {
            this._monitoring_interval = 3000;
         }
      }

      this._monitoring_on = monitor_on;

      if (this.isMonitoring())
         this._runMonitoring();
   }

   /** @summary Runs monitoring event loop
     * @private */
   _runMonitoring(arg) {
      if ((arg == "cleanup") || !this.isMonitoring()) {
         if (this._monitoring_handle) {
            clearTimeout(this._monitoring_handle);
            delete this._monitoring_handle;
         }

         if (this._monitoring_frame) {
            cancelAnimationFrame(this._monitoring_frame);
            delete this._monitoring_frame;
         }
         return;
      }

      if (arg == "frame") {
         // process of timeout, request animation frame
         delete this._monitoring_handle;
         this._monitoring_frame = requestAnimationFrame(this._runMonitoring.bind(this,"draw"));
         return;
      }

      if (arg == "draw") {
         delete this._monitoring_frame;
         this.updateItems();
      }

      this._monitoring_handle = setTimeout(this._runMonitoring.bind(this,"frame"), this.getMonitoringInterval());
   }

   /** @summary Returns configured monitoring interval in ms */
   getMonitoringInterval() {
      return this._monitoring_interval || 3000;
   }

   /** @summary Returns true when monitoring is enabled */
   isMonitoring() {
      return this._monitoring_on;
   }

   /** @summary Assign default layout and place where drawing will be performed
     * @param {string} layout - layout like "simple" or "grid2x2"
     * @param {string} frameid - DOM element id where object drawing will be performed */
   setDisplay(layout, frameid) {
      if (!frameid && (typeof layout == 'object')) {
         this.disp = layout;
         this.disp_kind = 'custom';
         this.disp_frameid = null;
      } else {
         this.disp_kind = layout;
         this.disp_frameid = frameid;
      }

      if (!this.register_resize && (this.disp_kind != 'batch')) {
         this.register_resize = true;
         registerForResize(this);
      }
   }

   /** @summary Returns configured layout */
   getLayout() {
      return this.disp_kind;
   }

   /** @summary Remove painter reference from hierarhcy
     * @private */
   removePainter(obj_painter) {
      this.forEachItem(item => {
         if (item._painter === obj_painter) {
            // delete painter reference
            delete item._painter;
            // also clear data which could be associated with item
            if (typeof item.clear == 'function') item.clear();
         }
      });
   }

   /** @summary Cleanup all items in hierarchy
     * @private */
   clearHierarchy(withbrowser) {
      if (this.disp) {
         this.disp.cleanup();
         delete this.disp;
      }

      let plainarr = [];

      this.forEachItem(item => {
         delete item._painter; // remove reference on the painter
         // when only display cleared, try to clear all browser items
         if (!withbrowser && (typeof item.clear == 'function')) item.clear();
         if (withbrowser) plainarr.push(item);
      });

      if (withbrowser) {
         // cleanup all monitoring loops
         this.enableMonitoring(false);
         // simplify work for javascript and delete all (ok, most of) cross-references
         this.selectDom().html("");
         plainarr.forEach(d => { delete d._parent; delete d._childs; delete d._obj; delete d._d3cont; });
         delete this.h;
      }
   }

   /** @summary Returns actual MDI display object
     * @desc It should an instance of {@link MDIDsiplay} class */
   getDisplay() {
      return this.disp;
   }

   /** @summary method called when MDI element is cleaned up
     * @desc hook to perform extra actions when frame is cleaned
     * @private */
   cleanupFrame(frame) {

      d3_select(frame).attr("frame_title", null);

      this.clearDrop(frame);

      let lst = cleanup(frame);

      // we remove all painters references from items
      if (lst.length > 0)
         this.forEachItem(item => {
            if (item._painter && lst.indexOf(item._painter) >= 0)
               delete item._painter;
         });
   }

   /** @summary Creates configured MDIDisplay object
     * @returns {Promise} when ready
     * @private */
   createDisplay() {

      if ('disp' in this) {
         if ((this.disp.numDraw() > 0) || (this.disp_kind == "custom"))
            return Promise.resolve(this.disp);
         this.disp.cleanup();
         delete this.disp;
      }

      if (this.disp_kind == 'batch') {
         let pr = isNodeJs() ? loadJSDOM() : Promise.resolve(null);
         return pr.then(handle => {
            this.disp = new BatchDisplay(1200, 800, handle?.body);
            return this.disp;
         });
      }

      // check that we can found frame where drawing should be done
      if (!document.getElementById(this.disp_frameid))
         return Promise.resolve(null);

      if ((this.disp_kind.indexOf("flex") == 0) || (this.disp_kind == "tabs") || (this.disp_kind.indexOf("coll") == 0))
         this.disp = new FlexibleDisplay(this.disp_frameid);
      else
         this.disp = new GridDisplay(this.disp_frameid, this.disp_kind);

      this.disp.cleanupFrame = this.cleanupFrame.bind(this);
      if (settings.DragAndDrop)
          this.disp.setInitFrame(this.enableDrop.bind(this));

      return Promise.resolve(this.disp);
   }

   /** @summary If possible, creates custom MDIDisplay for given item
     * @param itemname - name of item, for which drawing is created
     * @param custom_kind - display kind
     * @returns {Promise} with mdi object created
     * @private */
   createCustomDisplay(itemname, custom_kind) {

      if (this.disp_kind != "simple")
         return this.createDisplay();

      this.disp_kind = custom_kind;

      // check if display can be erased
      if (this.disp) {
         let num = this.disp.numDraw();
         if ((num > 1) || ((num == 1) && !this.disp.findFrame(itemname)))
            return this.createDisplay();
         this.disp.cleanup();
         delete this.disp;
      }

      return this.createDisplay();
   }

   /** @summary function updates object drawings for other painters
     * @private */
   updateOnOtherFrames(painter, obj) {
      let mdi = this.disp, isany = false;
      if (!mdi) return false;

      let handle = obj._typename ? getDrawHandle("ROOT." + obj._typename) : null;
      if (handle && handle.draw_field && obj[handle.draw_field])
         obj = obj[handle.draw_field];

      mdi.forEachPainter((p, frame) => {
         if ((p === painter) || (p.getItemName() != painter.getItemName())) return;

         // do not actiavte frame when doing update
         // mdi.activateFrame(frame);
         if ((typeof p.redrawObject == 'function') && p.redrawObject(obj)) isany = true;
      });
      return isany;
   }

   /** @summary Process resize event
     * @private */
   checkResize(size) {
      if (this.disp) this.disp.checkMDIResize(null, size);
   }

   /** @summary Load and execute scripts, kept to support v6 applications
     * @private */
   loadScripts(scripts, modules) {
      if (!scripts && !modules)
         return Promise.resolve(true);

      return _ensureJSROOT().then(v6 => {
         return v6.require(modules)
                  .then(() => loadScript(scripts))
                  .then(() => v6._complete_loading());
      });
   }

   /** @summary Start GUI
     * @returns {Promise} when ready
     * @private */
   startGUI(gui_div, url) {

      let d = decodeUrl(url);

      let GetOption = opt => {
         let res = d.get(opt, null);
         if ((res===null) && gui_div && !gui_div.empty() && gui_div.node().hasAttribute(opt))
            res = gui_div.attr(opt);
         return res;
      };

      let GetUrlOptionAsArray = opt => {

         let res = [];

         while (opt.length > 0) {
            let separ = opt.indexOf(";");
            let part = (separ>0) ? opt.substr(0, separ) : opt;

            if (separ > 0) opt = opt.substr(separ+1); else opt = "";

            let canarray = true;
            if (part[0]=='#') { part = part.substr(1); canarray = false; }

            let val = d.get(part,null);

            if (canarray)
               res = res.concat(parseAsArray(val));
            else if (val!==null)
               res.push(val);
         }
         return res;
      };

      let GetOptionAsArray = opt => {
         let res = GetUrlOptionAsArray(opt);
         if (res.length>0 || !gui_div || gui_div.empty()) return res;
         while (opt.length>0) {
            let separ = opt.indexOf(";");
            let part = separ>0 ? opt.substr(0, separ) : opt;
            if (separ>0) opt = opt.substr(separ+1); else opt = "";

            let canarray = true;
            if (part[0]=='#') { part = part.substr(1); canarray = false; }
            if (part==='files') continue; // special case for normal UI

            if (!gui_div.node().hasAttribute(part)) continue;

            let val = gui_div.attr(part);

            if (canarray) res = res.concat(parseAsArray(val));
            else if (val!==null) res.push(val);
         }
         return res;
      };

      let prereq = GetOption('prereq') || "",
          filesdir = d.get("path") || "", // path used in normal gui
          filesarr = GetOptionAsArray("#file;files"),
          localfile = GetOption("localfile"),
          jsonarr = GetOptionAsArray("#json;jsons"),
          expanditems = GetOptionAsArray("expand"),
          focusitem = GetOption("focus"),
          itemsarr = GetOptionAsArray("#item;items"),
          optionsarr = GetOptionAsArray("#opt;opts"),
          monitor = GetOption("monitoring"),
          layout = GetOption("layout"),
          style = GetOptionAsArray("#style"),
          statush = 0, status = GetOption("status"),
          browser_kind = GetOption("browser"),
          browser_configured = !!browser_kind,
          title = GetOption("title");

      if (monitor === null)
         monitor = 0;
      else if (monitor === "")
         monitor = 3000;
      else
         monitor = parseInt(monitor);

      if (GetOption("float")!==null) { browser_kind = 'float'; browser_configured = true; } else
      if (GetOption("fix")!==null) { browser_kind = 'fix'; browser_configured = true; }

      this.no_select = GetOption("noselect");

      if (GetOption('files_monitoring')!==null) this.files_monitoring = true;

      if (title) document.title = title;

      let load = GetOption("load");

      if (expanditems.length==0 && (GetOption("expand")==="")) expanditems.push("");

      if (filesdir) {
         for (let i=0;i<filesarr.length;++i) filesarr[i] = filesdir + filesarr[i];
         for (let i=0;i<jsonarr.length;++i) jsonarr[i] = filesdir + jsonarr[i];
      }

      if ((itemsarr.length == 0) && GetOption("item") === "") itemsarr.push("");

      if ((jsonarr.length == 1) && (itemsarr.length == 0) && (expanditems.length==0)) itemsarr.push("");

      if (!this.disp_kind) {
         if ((typeof layout == "string") && (layout.length > 0))
            this.disp_kind = layout;
         else
         switch (itemsarr.length) {
           case 0:
           case 1: this.disp_kind = 'simple'; break;
           case 2: this.disp_kind = 'vert2'; break;
           case 3: this.disp_kind = 'vert21'; break;
           case 4: this.disp_kind = 'vert22'; break;
           case 5: this.disp_kind = 'vert32'; break;
           case 6: this.disp_kind = 'vert222'; break;
           case 7: this.disp_kind = 'vert322'; break;
           case 8: this.disp_kind = 'vert332'; break;
           case 9: this.disp_kind = 'vert333'; break;
           default: this.disp_kind = 'flex';
         }
      }

      if (status==="no")
         status = null;
      else if (status==="off") {
         this.status_disabled = true;
         status = null;
      } else if (status==="on")
         status = true;
      else if (status!==null) {
         statush = parseInt(status);
         if (!Number.isInteger(statush) || (statush < 5)) statush = 0;
         status = true;
      }
      if (this.no_select === "") this.no_select = true;

      if (!browser_kind)
         browser_kind = "fix";
      else if (browser_kind === "no")
         browser_kind = "";
      else if (browser_kind==="off") {
         browser_kind = "";
         status = null;
         this.exclude_browser = true;
      }
      if (GetOption("nofloat")!==null) this.float_browser_disabled = true;

      if (this.start_without_browser) browser_kind = "";

      this._topname = GetOption("topname");

      let openAllFiles = () => {
         let promise;

         if (load || prereq) {
            promise = this.loadScripts(load, prereq); load = ""; prereq = "";
         } else if (browser_kind) {
            promise = this.createBrowser(browser_kind); browser_kind = "";
         } else if (status !== null) {
            promise = this.createStatusLine(statush, status); status = null;
         } else if (jsonarr.length > 0)
            promise = this.openJsonFile(jsonarr.shift());
         else if (filesarr.length > 0)
            promise = this.openRootFile(filesarr.shift());
         else if ((localfile!==null) && (typeof this.selectLocalFile == 'function')) {
            localfile = null; promise = this.selectLocalFile();
         } else if (expanditems.length > 0)
            promise = this.expandItem(expanditems.shift());
         else if (style.length > 0)
            promise = this.applyStyle(style.shift());
         else
            return this.refreshHtml()
                   .then(() => this.displayItems(itemsarr, optionsarr))
                   .then(() => focusitem ? this.focusOnItem(focusitem) : this)
                   .then(() => {
                      this.setMonitoring(monitor);
                      return itemsarr ? this.refreshHtml() : this; // this is final return
                   });

         return promise.then(openAllFiles);
      };

      let h0 = null;
      if (this.is_online) {
         let func = internals.GetCachedHierarchy || findFunction('GetCachedHierarchy');
         if (typeof func == 'function')
            h0 = func();
         if (typeof h0 !== 'object') h0 = "";

         if ((this.is_online == "draw") && !itemsarr.length)
            itemsarr.push("");
      }

      if (h0 !== null)
         return this.openOnline(h0).then(() => {
            // check if server enables monitoring
            if (!this.exclude_browser && !browser_configured && ('_browser' in this.h)) {
               browser_kind = this.h._browser;
               if (browser_kind==="no") browser_kind = ""; else
               if (browser_kind==="off") { browser_kind = ""; status = null; this.exclude_browser = true; }
            }

            if (('_monitoring' in this.h) && !monitor)
               monitor = this.h._monitoring;

            if (('_loadfile' in this.h) && (filesarr.length==0))
               filesarr = parseAsArray(this.h._loadfile);

            if (('_drawitem' in this.h) && (itemsarr.length==0)) {
               itemsarr = parseAsArray(this.h._drawitem);
               optionsarr = parseAsArray(this.h._drawopt);
            }

            if (('_layout' in this.h) && !layout && ((this.is_online != "draw") || (itemsarr.length > 1)))
               this.disp_kind = this.h._layout;

            if (('_toptitle' in this.h) && this.exclude_browser && document)
               document.title = this.h._toptitle;

            if (gui_div)
               this.prepareGuiDiv(gui_div.attr('id'), this.disp_kind);

            return openAllFiles();
         });

      if (gui_div)
         this.prepareGuiDiv(gui_div.attr('id'), this.disp_kind);

      return openAllFiles();
   }

   /** @summary Prepare div element - create layout and buttons
     * @private */
   prepareGuiDiv(myDiv, layout) {

      this.gui_div = (typeof myDiv == "string") ? myDiv : myDiv.attr('id');

      this.brlayout = new BrowserLayout(this.gui_div, this);

      this.brlayout.create(!this.exclude_browser);

      if (!this.exclude_browser) {
         let btns = this.brlayout.createBrowserBtns();

         ToolbarIcons.createSVG(btns, ToolbarIcons.diamand, 15, "toggle fix-pos browser")
                     .style("margin","3px").on("click", () => this.createBrowser("fix", true));

         if (!this.float_browser_disabled)
            ToolbarIcons.createSVG(btns, ToolbarIcons.circle, 15, "toggle float browser")
                        .style("margin","3px").on("click", () => this.createBrowser("float", true));

         if (!this.status_disabled)
            ToolbarIcons.createSVG(btns, ToolbarIcons.three_circles, 15, "toggle status line")
                        .style("margin","3px").on("click", () => this.createStatusLine(0, "toggle"));
      }

      this.setDisplay(layout, this.brlayout.drawing_divid());
   }

   /** @summary Create status line
     * @param {number} [height] - size of the status line
     * @param [mode] - false / true / "toggle"
     * @returns {Promise} when ready */
   createStatusLine(height, mode) {
      if (this.status_disabled || !this.gui_div || !this.brlayout)
         return Promise.resolve("");
      return this.brlayout.createStatusLine(height, mode);
   }

   /** @summary Redraw hierarchy
     * @desc works only when inspector or streamer info is displayed
     * @private */
   redrawObject(obj) {
      if (!this._inspector && !this._streamer_info) return false;
      if (this._streamer_info)
         this.h = createStreamerInfoContent(obj)
      else
         this.h = createInspectorContent(obj);
      return this.refreshHtml().then(() => { this.setTopPainter(); });
   }

   /** @summary Create browser elements
     * @returns {Promise} when completed */
   createBrowser(browser_kind, update_html) {

      if (!this.gui_div || this.exclude_browser || !this.brlayout)
         return Promise.resolve(false);

      let main = d3_select("#" + this.gui_div + " .jsroot_browser");

      // one requires top-level container
      if (main.empty())
         return Promise.resolve(false);

      if ((browser_kind==="float") && this.float_browser_disabled) browser_kind = "fix";

      if (!main.select('.jsroot_browser_area').empty()) {
         // this is case when browser created,
         // if update_html specified, hidden state will be toggled

         if (update_html) this.brlayout.toggleKind(browser_kind);

         return Promise.resolve(true);
      }

      let guiCode = `<p class="jsroot_browser_version"><a href="https://root.cern/js/">JSROOT</a> version <span style="color:green"><b>${version}</b></span></p>`;

      if (this.is_online) {
         guiCode += '<p> Hierarchy in <a href="h.json">json</a> and <a href="h.xml">xml</a> format</p>' +
                    '<div style="display:inline; vertical-align:middle; white-space: nowrap;">' +
                    '<label style="margin-right:5px"><input type="checkbox" name="monitoring" class="gui_monitoring"/>Monitoring</label>';
      } else if (!this.no_select) {
         let myDiv = d3_select("#"+this.gui_div),
             files = myDiv.attr("files") || "../files/hsimple.root",
             path = decodeUrl().get("path") || myDiv.attr("path") || "",
             arrFiles = files.split(';');

         guiCode += '<input type="text" value="" style="width:95%; margin:5px;border:2px;" class="gui_urlToLoad" title="input file name"/>' +
                    '<div style="display:flex;flex-direction:row;padding-top:5px">' +
                    '<select class="gui_selectFileName" style="flex:1;padding:2px;" title="select file name"' +
                    '<option value="" selected="selected"></option>';
         arrFiles.forEach(fname => { guiCode += `<option value="${path + fname}">${fname}</option>`; });
         guiCode += '</select>' +
                    '<input type="file" class="gui_localFile" accept=".root" style="display:none"/><output id="list" style="display:none"></output>' +
                    '<input type="button" value="..." class="gui_fileBtn" style="min-width:3em;padding:3px;margin-left:5px;margin-right:5px;" title="select local file for reading"/><br/>' +
                    '</div>' +
                    '<p id="gui_fileCORS"><small><a href="https://github.com/root-project/jsroot/blob/master/docs/JSROOT.md#reading-root-files-from-other-servers">Read docu</a>' +
                    ' how to open files from other servers.</small></p>' +
                    '<div style="display:flex;flex-direction:row">' +
                    '<input style="padding:3px;margin-right:5px" class="gui_ReadFileBtn" type="button" title="Read the Selected File" value="Load"/>' +
                    '<input style="padding:3px;margin-right:5px" class="gui_ResetUIBtn" type="button" title="Close all opened files and clear drawings" value="Reset"/>';
      } else if (this.no_select == "file") {
         guiCode += '<div style="display:flex;flex-direction:row">';
      }

      if (this.is_online || !this.no_select || this.no_select=="file")
         guiCode += '<select style="padding:2px;margin-right:5px;" title="layout kind" class="gui_layout"></select>' +
                    '</div>';

      guiCode += `<div id="${this.gui_div}_browser_hierarchy" class="jsroot_browser_hierarchy"></div>`;

      this.brlayout.setBrowserContent(guiCode);

      if (this.is_online)
          this.brlayout.setBrowserTitle('ROOT online server');
       else
          this.brlayout.setBrowserTitle('Read a ROOT file');

      let localfile_read_callback = null;

      if (!this.is_online && !this.no_select) {

         this.readSelectedFile = function() {
            let filename = main.select(".gui_urlToLoad").property('value').trim();
            if (!filename) return;

            if (filename.toLowerCase().lastIndexOf(".json") == filename.length-5)
               this.openJsonFile(filename);
            else
               this.openRootFile(filename);
         };

         main.select(".gui_selectFileName").property("value", "")
              .on("change", evnt => main.select(".gui_urlToLoad").property("value", evnt.target.value));
         main.select(".gui_fileBtn").on("click", () => main.select(".gui_localFile").node().click());

         main.select(".gui_ReadFileBtn").on("click", () => this.readSelectedFile());

         main.select(".gui_ResetUIBtn").on("click", () => this.clearHierarchy(true));

         main.select(".gui_urlToLoad").on("keyup", evnt => {
            if (evnt.keyCode == 13) this.readSelectedFile();
         });

         main.select(".gui_localFile").on('change', evnt => {
            let files = evnt.target.files, promises = [];

            for (let n = 0; n < files.length; ++n) {
               let f = files[n];
               main.select(".gui_urlToLoad").property('value', f.name);
               promises.push(this.openRootFile(f));
            }

            Promise.all(promises).then(() => {
               if (localfile_read_callback) {
                  localfile_read_callback();
                  localfile_read_callback = null;
               }
            });
         });

         this.selectLocalFile = function() {
            return new Promise(resolveFunc => {
               localfile_read_callback = resolveFunc;
               main.select(".gui_localFile").node().click();
            });
         };
      }

      let layout = main.select(".gui_layout");
      if (!layout.empty()) {
         ['simple', 'vert2', 'vert3', 'vert231', 'horiz2', 'horiz32', 'flex',
          'grid 2x2', 'grid 1x3', 'grid 2x3', 'grid 3x3', 'grid 4x4'].forEach(kind => layout.append("option").attr("value", kind).html(kind));

         layout.on('change', ev => this.setDisplay(ev.target.value || 'flex', this.gui_div + "_drawing"));
      }

      this.setDom(this.gui_div + '_browser_hierarchy');

      if (update_html) {
         this.refreshHtml();
         this.initializeBrowser();
      }

      return this.brlayout.toggleBrowserKind(browser_kind || "fix");
   }

   /** @summary Initialize browser elements */
   initializeBrowser() {

      let main = d3_select("#" + this.gui_div + " .jsroot_browser");
      if (main.empty() || !this.brlayout) return;

      if (this.brlayout) this.brlayout.adjustBrowserSize();

      let selects = main.select(".gui_layout").node();

      if (selects) {
         let found = false;
         for (let i in selects.options) {
            let s = selects.options[i].text;
            if (typeof s !== 'string') continue;
            if ((s == this.getLayout()) || (s.replace(/ /g,"") == this.getLayout())) {
               selects.selectedIndex = i; found = true;
               break;
            }
         }
         if (!found) {
            let opt = document.createElement('option');
            opt.innerHTML = opt.value = this.getLayout();
            selects.appendChild(opt);
            selects.selectedIndex = selects.options.length-1;
         }
      }

      if (this.is_online) {
         if (this.h && this.h._toptitle)
            this.brlayout.setBrowserTitle(this.h._toptitle);
         main.select(".gui_monitoring")
           .property('checked', this.isMonitoring())
           .on("click", evnt => {
               this.enableMonitoring(evnt.target.checked);
               this.updateItems();
            });
      } else if (!this.no_select) {
         let fname = "";
         this.forEachRootFile(item => { if (!fname) fname = item._fullurl; });
         main.select(".gui_urlToLoad").property("value", fname);
      }
   }

   /** @summary Enable monitoring mode */
   enableMonitoring(on) {
      this.setMonitoring(undefined, on);

      let chkbox = d3_select("#" + this.gui_div + " .jsroot_browser .gui_monitoring");
      if (!chkbox.empty() && (chkbox.property('checked') !== on))
         chkbox.property('checked', on);
   }

} // class HierarchyPainter



/** @summary Show object in inspector for provided object
  * @protected */
ObjectPainter.prototype.showInspector = function(obj) {
   if (obj === 'check')
      return true;

   let main = this.selectDom(),
      rect = getElementRect(main),
      w = Math.round(rect.width * 0.05) + "px",
      h = Math.round(rect.height * 0.05) + "px",
      id = "root_inspector_" + internals.id_counter++;

   main.append("div")
      .attr("id", id)
      .attr("class", "jsroot_inspector")
      .style('position', 'absolute')
      .style('top', h)
      .style('bottom', h)
      .style('left', w)
      .style('right', w);

   if (!obj || (typeof obj !== 'object') || !obj._typename)
      obj = this.getObject();

   return drawInspector(id, obj);
}





/** @summary Display streamer info
  * @private */
function drawStreamerInfo(dom, lst) {
   let painter = new HierarchyPainter('sinfo', dom, 'white');

   // in batch mode HTML drawing is not possible, just keep object reference for a minute
   if (isBatchMode()) {
      painter.selectDom().property("_json_object_", lst);
      return Promise.resolve(painter);
   }

   painter._streamer_info = true;
   painter.h = createStreamerInfoContent(lst);

   // painter.selectDom().style('overflow','auto');

   return painter.refreshHtml().then(() => {
      painter.setTopPainter();
      return painter;
   });
}

// ======================================================================================

/** @summary Display inspector
  * @private */
function drawInspector(dom, obj) {

   cleanup(dom);
   let painter = new HierarchyPainter('inspector', dom, 'white');

   // in batch mode HTML drawing is not possible, just keep object reference for a minute
   if (isBatchMode()) {
      painter.selectDom().property("_json_object_", obj);
      return Promise.resolve(painter);
   }

   painter.default_by_click = "expand"; // default action
   painter.with_icons = false;
   painter._inspector = true; // keep

   if (painter.selectDom().classed("jsroot_inspector"))
      painter.removeInspector = function() {
         this.selectDom().remove();
      }

   painter.fill_context = function(menu, hitem) {
      let sett = getDrawSettings(hitem._kind, 'nosame');
      if (sett.opts)
         menu.addDrawMenu("nosub:Draw", sett.opts, function(arg) {
            if (!hitem || !hitem._obj) return;
            let obj = hitem._obj, ddom = this.selectDom().node();
            if (this.removeInspector) {
               ddom = ddom.parentNode;
               this.removeInspector();
               if (arg == "inspect")
                  return this.showInspector(obj);
            }
            cleanup(ddom);
            draw(ddom, obj, arg);
         });
   }

   painter.h = createInspectorContent(obj);

   return painter.refreshHtml().then(() => {
      painter.setTopPainter();
      return painter;
   });
}


export { getHPainter, HierarchyPainter,
         drawInspector, drawStreamerInfo, drawList,
         folderHierarchy, taskHierarchy, listHierarchy, objectHierarchy, keysHierarchy };
